# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import Literal

import logging
from collections.abc import Callable
from pathlib import Path

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Soup

import nbxmpp
from nbxmpp.const import HTTPRequestError

log = logging.getLogger("nbxmpp.http")


HTTP_METHODS_T = Literal[
    "CONNECT",
    "DELETE",
    "GET",
    "HEAD",
    "OPTIONS",
    "PATCH",
    "POST",
    "PUT",
    "TRACE",
]
CHUNK_SIZE = 32768
DEFAULT_USER_AGENT = f"nbxmpp/{nbxmpp.__version__}"
SIGNAL_ACTIONS = GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION
MIN_SOUP_3_3_0 = Soup.check_version(3, 3, 0)


class HTTPLogAdapter(logging.LoggerAdapter):
    def process(self, msg: str, kwargs: Any) -> tuple[str, Any]:
        return f'{self.extra["request"]}: {msg}', kwargs


class HTTPSession:
    def __init__(self, user_agent: str = DEFAULT_USER_AGENT) -> None:
        self._session = Soup.Session()
        self._session.set_user_agent(user_agent)
        self._session.add_feature_by_type(Soup.ContentSniffer)
        self._session.remove_feature_by_type(Soup.ContentDecoder)

    def get_soup_session(self) -> Soup.Session:
        return self._session

    def set_proxy_resolver(self, resolver: Gio.SimpleProxyResolver | None) -> None:

        self._session.set_proxy_resolver(resolver)

    def create_request(self, force_http1: bool = False) -> HTTPRequest:
        return HTTPRequest(self, force_http1=force_http1)


class HTTPRequest(GObject.GObject):

    __gtype_name__ = "HTTPRequest"

    __gsignals__ = {
        "content-sniffed": (SIGNAL_ACTIONS, None, (int, str)),
        "starting-response-body": (SIGNAL_ACTIONS, None, ()),
        "response-progress": (SIGNAL_ACTIONS, None, (float,)),
        "request-progress": (SIGNAL_ACTIONS, None, (float,)),
        "finished": (SIGNAL_ACTIONS, None, ()),
        "destroy": (SIGNAL_ACTIONS, None, ()),
    }

    def __init__(self, session: HTTPSession, force_http1: bool = False) -> None:
        GObject.GObject.__init__(self)

        self._log = HTTPLogAdapter(log, extra={"request": self})

        self._session = session

        self._received_size = 0
        self._sent_size = 0

        self._cancellable = Gio.Cancellable()
        self._input_stream: Gio.InputStream | None = None
        self._output_stream: Gio.OutputStream | None = None
        self._is_finished = False
        self._error: HTTPRequestError | None = None
        self._is_complete = False
        self._timeout_reached = False
        self._timeout_id: int | None = None
        self._accept_certificate_func = None

        self._response_body_file: Gio.File | None = None
        self._response_body_data = b""
        self._body_received = False

        self._request_body_file: Gio.File | None = None
        self._request_body_data: bytes | None = None

        self._request_content_type = ""
        self._request_content_length = 0

        self._response_content_type = ""
        self._response_content_length = 0

        self._emit_request_progress = False
        self._emit_response_progress = False

        self._message = Soup.Message()
        if MIN_SOUP_3_3_0:
            self._message.set_force_http1(force_http1)  # type: ignore
        self._user_data = None

        self._log.info("Created")

    def is_finished(self) -> bool:
        return self._is_finished

    def is_complete(self) -> bool:
        return self._is_complete

    def get_error(self) -> HTTPRequestError | None:
        return self._error

    def get_error_string(self) -> str:
        if self._error == HTTPRequestError.STATUS_NOT_OK:
            return self._message.get_reason_phrase()
        return repr(self._error)

    def get_response_headers(self) -> Soup.MessageHeaders:
        return self._message.get_response_headers()

    def get_request_headers(self) -> Soup.MessageHeaders:
        return self._message.get_request_headers()

    def get_data(self) -> bytes:
        if not self._is_finished:
            raise ValueError("Process not finished, data not available")
        return self._response_body_data

    def get_uri(self) -> GLib.Uri | None:
        return self._message.get_uri()

    def get_status(self) -> Soup.Status:
        return self._message.get_status()

    def get_soup_message(self) -> Soup.Message:
        return self._message

    def set_user_data(self, user_data: Any) -> None:
        self._user_data = user_data

    def get_user_data(self) -> Any:
        return self._user_data

    def cancel(self) -> None:
        if self._is_finished:
            raise ValueError("Session already finished")

        self._log.info("Cancel requested")
        self._cancellable.cancel()

    def set_accept_certificate_func(self, func: Any) -> None:
        self._accept_certificate_func = func

    def set_request_body_from_path(self, content_type: str, path: Path) -> None:
        if not path.exists():
            raise ValueError("%s does not exist" % path)

        if not path.is_file():
            raise ValueError("%s is not a file" % path)

        self._request_body_file = Gio.File.new_for_path(str(path))
        self._request_content_type = content_type
        self._request_content_length = path.stat().st_size

    def set_request_body(self, content_type: str, data: bytes) -> None:
        self._request_body_data = data
        self._request_content_type = content_type
        self._request_content_length = len(data)

    def set_response_body_from_path(self, path: Path) -> None:
        self._response_body_file = Gio.File.new_for_path(str(path))

    def connect(self, signal_name: str, callback: Any, *args: Any) -> None:

        if signal_name == "response-progress":
            self._emit_response_progress = True

        if signal_name == "request-progress":
            self._emit_request_progress = True

        if signal_name in GObject.signal_list_names(HTTPRequest):
            GObject.GObject.connect(self, signal_name, callback, *args)
            return

        user_data = (callback, args)

        self._message.connect(signal_name, self._on_connect_callback, user_data)

    def _on_connect_callback(self, _message: Soup.Message, *args: Any) -> None:
        signal_args = args[:-1]
        callback, user_data = args[-1]
        callback(self, *signal_args, *user_data)

    def send(
        self,
        method: HTTP_METHODS_T,
        uri_string: str,
        timeout: int | None = None,
        callback: Callable[[HTTPRequest], Any] | None = None,
    ) -> None:

        if callback:
            self.connect("finished", callback)
        self._send(method, uri_string, timeout)

    def _send(
        self, method: HTTP_METHODS_T, uri_string: str, timeout: int | None = None
    ) -> None:

        if self._is_finished:
            raise ValueError("Session already finished")

        self._message.set_method(method)

        if self._response_body_file is not None:
            self._output_stream = self._response_body_file.replace(
                None, False, Gio.FileCreateFlags.REPLACE_DESTINATION, self._cancellable
            )

        uri = GLib.Uri.parse(uri_string, GLib.UriFlags(Soup.HTTP_URI_FLAGS))
        self._message.set_uri(uri)

        if self._request_body_data is not None:
            self._message.set_request_body_from_bytes(
                self._request_content_type, GLib.Bytes.new(self._request_body_data)
            )
            if self._emit_request_progress:
                self._message.connect("wrote-body-data", self._on_request_body_progress)

        if self._request_body_file is not None:
            request_input_stream = self._request_body_file.read(self._cancellable)
            self._message.set_request_body(
                self._request_content_type,
                request_input_stream,
                self._request_content_length,
            )
            if self._emit_request_progress:
                self._message.connect("wrote-body-data", self._on_request_body_progress)

        self._message.connect("content-sniffed", self._on_content_sniffed)
        self._message.connect("got-body", self._on_got_body)
        self._message.connect("restarted", self._on_restarted)
        self._message.connect("finished", self._on_finished)
        self._message.connect("got-headers", self._on_got_headers)
        if self._accept_certificate_func is not None:
            self._message.connect("accept-certificate", self._accept_certificate_func)

        soup_session = self._session.get_soup_session()
        soup_session.send_async(
            self._message, GLib.PRIORITY_DEFAULT, self._cancellable, self._on_response
        )

        if timeout is not None:
            self._timeout_id = GLib.timeout_add_seconds(timeout, self._on_timeout)

        self._log.info("Request sent, method: %s, uri: %s", method, uri_string)

    def _on_request_body_progress(self, _message: Soup.Message, chunk: int) -> None:

        self._sent_size += chunk
        self.emit("request-progress", self._sent_size / self._request_content_length)

    def _on_timeout(self) -> None:
        self._timeout_reached = True
        self._timeout_id = None
        self._set_error(HTTPRequestError.TIMEOUT)
        self.cancel()

    def _on_response(self, session: Soup.Session, result: Gio.AsyncResult) -> None:

        self._log.info("HTTP version: %s", self._message.get_http_version().value_name)
        self._log.info("Request response received")
        try:
            self._input_stream = session.send_finish(result)
        except GLib.Error as error:
            quark = GLib.quark_try_string("g-io-error-quark")
            if error.matches(quark, Gio.IOErrorEnum.CANCELLED):
                self._set_failed(HTTPRequestError.CANCELLED)
                return

            self._log.error(error)
            self._set_failed(HTTPRequestError.UNKNOWN)
            return

        status = self._message.get_status()
        if status not in (Soup.Status.OK, Soup.Status.CREATED):
            self._log.info(
                "Response status: %s %s", int(status), self._message.get_reason_phrase()
            )
            self._set_failed(HTTPRequestError.STATUS_NOT_OK)
            return

        self._log.info("Start downloading response body")
        self.emit("starting-response-body")

        self._read_async()

    def _read_async(self) -> None:
        assert self._input_stream is not None
        self._input_stream.read_bytes_async(
            CHUNK_SIZE, GLib.PRIORITY_LOW, self._cancellable, self._on_bytes_read_result
        )

    def _on_bytes_read_result(
        self, input_stream: Gio.InputStream, result: Gio.AsyncResult
    ) -> None:

        try:
            data = input_stream.read_bytes_finish(result)
        except GLib.Error as error:
            self._log.error(error)
            quark = GLib.quark_try_string("g-io-error-quark")
            if error.matches(quark, Gio.IOErrorEnum.CANCELLED):
                self._finish_read(HTTPRequestError.CANCELLED)
            else:
                self._finish_read(HTTPRequestError.UNKNOWN)

            return

        bytes_ = data.get_data()
        if not bytes_:
            self._finish_read()
            return

        self._received_size += len(bytes_)
        if self._check_content_overflow():
            return

        if self._output_stream is None:
            self._response_body_data += bytes_

        else:
            try:
                self._output_stream.write_all(bytes_, self._cancellable)
            except GLib.Error as error:
                self._log.error(error)
                quark = GLib.quark_try_string("g-io-error-quark")
                if error.matches(quark, Gio.IOErrorEnum.CANCELLED):
                    self._finish_read(HTTPRequestError.CANCELLED)
                else:
                    self._finish_read(HTTPRequestError.UNKNOWN)
                return

        self._read_async()
        self._emit_progress()

    def _finish_read(self, error: HTTPRequestError | None = None) -> None:
        self._log.info("Finished reading")
        if error is None:
            self._close_all_streams()
            return

        self._set_failed(error)

    def _on_content_sniffed(
        self,
        message: Soup.Message,
        content_type: str | None,
        _params: GLib.HashTable,
    ) -> None:

        # Signal is only raised when there is content in the response
        headers = message.get_response_headers()
        if content_type is None:
            # According to the docs, content_type is None when the sniffer
            # decides to trust the content-type sent by the server.
            content_type, _ = headers.get_content_type()

        self._response_content_type = content_type or ""

        self._log.info("Sniffed: content-type: %s", self._response_content_type)

        self.emit(
            "content-sniffed",
            self._response_content_length,
            self._response_content_type,
        )

    def _on_got_headers(self, message: Soup.Message) -> None:
        headers = message.get_response_headers()
        self._response_content_length = headers.get_content_length()
        self._log.info("Got Headers: content-length: %s", self._response_content_length)

    def _on_got_body(self, _message: Soup.Message) -> None:
        # This signal tells us that the full body was received.
        # The `finished` signal is not a sure indicator if the message body
        # was received in full, as its also triggered when a message is
        # cancelled.
        self._log.info("Body received")
        self._body_received = True

    def _emit_progress(self) -> None:
        if not self._emit_response_progress:
            return

        if not self._message.get_method() == "GET":
            return

        self.emit(
            "response-progress", self._received_size / self._response_content_length
        )

    def _check_content_overflow(self) -> bool:
        if self._received_size > self._response_content_length:
            self._finish_read(HTTPRequestError.CONTENT_OVERFLOW)
            return True
        return False

    def _on_restarted(self, _message: Soup.Message) -> None:
        self._log.info("Restarted")
        self._body_received = False
        self._response_content_type = ""
        self._response_content_length = 0
        self._received_size = 0

    def _on_finished(self, _message: Soup.Message) -> None:
        self._log.info("Message finished")
        if not self._body_received:
            # This can happen when the message is cancelled. The `finished`
            # signal is raised whenever the input stream is closed.
            # In the case the message was cancelled other parts of the code
            # will call set_failed().
            return

        if self._received_size != self._response_content_length:
            self._set_failed(HTTPRequestError.INCOMPLETE)
            return

        status = self._message.get_status()
        if status not in (Soup.Status.OK, Soup.Status.CREATED):
            self._set_failed(HTTPRequestError.STATUS_NOT_OK)
            return

        self._log.info("Request status: %s", Soup.Status.get_phrase(status))

        if self._cancellable.is_cancelled():
            # It can happen that the message is finished before the
            # response callback returns after calling cancel(). If
            # we call complete, the response callback will also
            # try to cleanup and will fail.
            self._log.info(
                "Skip setting message complete because " "cancel is in progess"
            )
            return

        self._set_complete()

    def _set_error(self, error: HTTPRequestError) -> None:
        self._log.info("Set Error: %s", error)
        self._error = error

    def _set_failed(self, error: HTTPRequestError) -> None:
        self._log.info("Set Failed: %s", error)
        self._is_finished = True

        if self._error is None:
            self._set_error(error)

        self._close_all_streams()
        self.emit("finished")
        self._cleanup()

    def _set_complete(self) -> None:
        self._log.info("Set Complete")
        self._is_finished = True
        self._is_complete = True
        self._close_all_streams()
        self.emit("finished")
        self._cleanup()

    def _close_all_streams(self) -> None:
        # stream.close() will invoke signals on the Message object
        # which in turn can lead to this method called again in the
        # same Mainloop iteration. This means is_closed() will not
        # return True and we get an GLib.IOError.PENDING error.

        input_stream = self._input_stream
        output_stream = self._output_stream

        self._input_stream = None
        self._output_stream = None

        if input_stream is not None:
            if not input_stream.is_closed():
                input_stream.close(None)

        if output_stream is not None:
            if not output_stream.is_closed():
                output_stream.close(None)

    def _cleanup(self) -> None:
        self._log.info("Run cleanup")

        self._response_body_data = b""
        self._message.run_dispose()

        del self._cancellable
        del self._user_data

        if self._timeout_id is not None:
            GLib.source_remove(self._timeout_id)
            self._timeout_id = None

        self.emit("destroy")
        self.run_dispose()

    def __repr__(self) -> str:
        return f"Request({id(self)})"



# Copyright (C) 2003-2005 Alexey "Snake" Nezhdanov
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Protocol module contains tools that are needed for processing of xmpp-related
data structures, including jabber-objects like JID or different stanzas and
sub- stanzas) handling routines
"""

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import TYPE_CHECKING

import functools
import hashlib
import os
import sqlite3
import time
import warnings
from base64 import b64encode
from collections.abc import Iterable
from dataclasses import asdict
from dataclasses import dataclass

import idna
from gi.repository import Gio
from gi.repository import GLib

from nbxmpp.namespaces import Namespace
from nbxmpp.precis import enforce_precis_opaque
from nbxmpp.precis import enforce_precis_username
from nbxmpp.simplexml import Node
from nbxmpp.stringprep import nodeprep
from nbxmpp.stringprep import resourceprep
from nbxmpp.xmppiri import clean_iri
from nbxmpp.xmppiri import escape_ifragment
from nbxmpp.xmppiri import escape_inode
from nbxmpp.xmppiri import escape_ires
from nbxmpp.xmppiri import escape_ivalue
from nbxmpp.xmppiri import validate_ikey
from nbxmpp.xmppiri import validate_querytype

if TYPE_CHECKING:
    from nbxmpp.structs import MucSubject


def ascii_upper(s: str) -> str:
    return s.upper()


SASL_AUTH_MECHS = [
    "SCRAM-SHA-512-PLUS",
    "SCRAM-SHA-256-PLUS",
    "SCRAM-SHA-1-PLUS",
    "SCRAM-SHA-512",
    "SCRAM-SHA-256",
    "SCRAM-SHA-1",
    "GSSAPI",
    "PLAIN",
    "EXTERNAL",
    "ANONYMOUS",
]

SASL_ERROR_CONDITIONS = [
    "aborted",
    "account-disabled",
    "credentials-expired",
    "encryption-required",
    "incorrect-encoding",
    "invalid-authzid",
    "invalid-mechanism",
    "mechanism-too-weak",
    "malformed-request",
    "not-authorized",
    "temporary-auth-failure",
]

ERRORS: dict[str, tuple[str, str]] = {
    "urn:ietf:params:xml:ns:xmpp-sasl aborted": (
        "",
        "The receiving entity acknowledges an <abort/> element sent by the initiating entity; sent in reply to the <abort/> element.",
    ),
    "urn:ietf:params:xml:ns:xmpp-sasl incorrect-encoding": (
        "",
        "The data provided by the initiating entity could not be processed because the [BASE64)Josefsson, S., The Base16, Base32, and Base64 Data Encodings, July 2003. encoding is incorrect (e.g., because the encoding does not adhere to the definition in Section 3 of [BASE64)Josefsson, S., The Base16, Base32, and Base64 Data Encodings, July 2003.); sent in reply to a <response/> element or an <auth/> element with initial response data.",
    ),
    "urn:ietf:params:xml:ns:xmpp-sasl invalid-authzid": (
        "",
        "The authzid provided by the initiating entity is invalid, either because it is incorrectly formatted or because the initiating entity does not have permissions to authorize that ID; sent in reply to a <response/> element or an <auth/> element with initial response data.",
    ),
    "urn:ietf:params:xml:ns:xmpp-sasl invalid-mechanism": (
        "",
        "The initiating entity did not provide a mechanism or requested a mechanism that is not supported by the receiving entity; sent in reply to an <auth/> element.",
    ),
    "urn:ietf:params:xml:ns:xmpp-sasl mechanism-too-weak": (
        "",
        "The mechanism requested by the initiating entity is weaker than server policy permits for that initiating entity; sent in reply to a <response/> element or an <auth/> element with initial response data.",
    ),
    "urn:ietf:params:xml:ns:xmpp-sasl not-authorized": (
        "",
        "The authentication failed because the initiating entity did not provide valid credentials (this includes but is not limited to the case of an unknown username); sent in reply to a <response/> element or an <auth/> element with initial response data.",
    ),
    "urn:ietf:params:xml:ns:xmpp-sasl temporary-auth-failure": (
        "",
        "The authentication failed because of a temporary error condition within the receiving entity; sent in reply to an <auth/> element or <response/> element.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas bad-request": (
        "modify",
        "The sender has sent XML that is malformed or that cannot be processed.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas conflict": (
        "cancel",
        "Access cannot be granted because an existing resource or session exists with the same name or address.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas feature-not-implemented": (
        "cancel",
        "The feature requested is not implemented by the recipient or server and therefore cannot be processed.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas forbidden": (
        "auth",
        "The requesting entity does not possess the required permissions to perform the action.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas gone": (
        "modify",
        "The recipient or server can no longer be contacted at this address.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas internal-server-error": (
        "wait",
        "The server could not process the stanza because of a misconfiguration or an otherwise-undefined internal server error.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas item-not-found": (
        "cancel",
        "The addressed JID or item requested cannot be found.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas jid-malformed": (
        "modify",
        "The value of the        'to' attribute in the sender's stanza does not adhere to the syntax defined in Addressing Scheme.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas not-acceptable": (
        "cancel",
        "The recipient or server understands the request but is refusing to process it because it does not meet criteria defined by the recipient or server.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas not-allowed": (
        "cancel",
        "The recipient or server does not allow any entity to perform the action.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas not-authorized": (
        "auth",
        "The sender must provide proper credentials before being allowed to perform the action, or has provided improper credentials.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas payment-required": (
        "auth",
        "The requesting entity is not authorized to access the requested service because payment is required.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas recipient-unavailable": (
        "wait",
        "The intended recipient is temporarily unavailable.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas redirect": (
        "modify",
        "The recipient or server is redirecting requests for this information to another entity.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas registration-required": (
        "auth",
        "The requesting entity is not authorized to access the requested service because registration is required.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas remote-server-not-found": (
        "cancel",
        "A remote server or service specified as part or all of the JID of the intended recipient does not exist.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas remote-server-timeout": (
        "wait",
        "A remote server or service specified as part or all of the JID of the intended recipient could not be contacted within a reasonable amount of time.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas resource-constraint": (
        "wait",
        "The server or recipient lacks the system resources necessary to service the request.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas service-unavailable": (
        "cancel",
        "The server or recipient does not currently provide the requested service.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas subscription-required": (
        "auth",
        "The requesting entity is not authorized to access the requested service because a subscription is required.",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas undefined-condition": (
        "",
        "Undefined Condition",
    ),
    "urn:ietf:params:xml:ns:xmpp-stanzas unexpected-request": (
        "wait",
        "The recipient or server understood the request but was not expecting it at this time (e.g., the request was out of order).",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams bad-format": (
        "",
        "The entity has sent XML that cannot be processed.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams bad-namespace-prefix": (
        "",
        "The entity has sent a namespace prefix that is unsupported, or has sent no namespace prefix on an element that requires such a prefix.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams conflict": (
        "",
        "The server is closing the active stream for this entity because a new stream has been initiated that conflicts with the existing stream.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams connection-timeout": (
        "",
        "The entity has not generated any traffic over the stream for some period of time.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams host-gone": (
        "",
        "The value of the        'to' attribute provided by the initiating entity in the stream header corresponds to a hostname that is no longer hosted by the server.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams host-unknown": (
        "",
        "The value of the        'to' attribute provided by the initiating entity in the stream header does not correspond to a hostname that is hosted by the server.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams improper-addressing": (
        "",
        "A stanza sent between two servers lacks a        'to' or 'from' attribute (or the attribute has no value).",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams internal-server-error": (
        "",
        "The server has experienced a misconfiguration or an otherwise-undefined internal error that prevents it from servicing the stream.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams invalid-from": (
        "",
        "The JID or hostname provided in a        'from' address does not match an authorized JID or validated domain negotiated between servers via SASL or dialback, or between a client and a server via authentication and resource authorization.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams invalid-id": (
        "",
        "The stream ID or dialback ID is invalid or does not match an ID previously provided.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams invalid-namespace": (
        "",
        'The streams namespace name is something other than        "http://etherx.jabber.org/streams" or the dialback namespace name is something other than "jabber:server:dialback".',
    ),
    "urn:ietf:params:xml:ns:xmpp-streams invalid-xml": (
        "",
        "The entity has sent invalid XML over the stream to a server that performs validation.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams not-authorized": (
        "",
        "The entity has attempted to send data before the stream has been authenticated, or otherwise is not authorized to perform an action related to stream negotiation.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams policy-violation": (
        "",
        "The entity has violated some local service policy.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams remote-connection-failed": (
        "",
        "The server is unable to properly connect to a remote resource that is required for authentication or authorization.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams resource-constraint": (
        "",
        "The server lacks the system resources necessary to service the stream.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams restricted-xml": (
        "",
        "The entity has attempted to send restricted XML features such as a comment, processing instruction, DTD, entity reference, or unescaped character.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams see-other-host": (
        "",
        "The server will not provide service to the initiating entity but is redirecting traffic to another host.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams system-shutdown": (
        "",
        "The server is being shut down and all active streams are being closed.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams undefined-condition": (
        "",
        "The error condition is not one of those defined by the other conditions in this list.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams unsupported-encoding": (
        "",
        "The initiating entity has encoded the stream in an encoding that is not supported by the server.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams unsupported-stanza-type": (
        "",
        "The initiating entity has sent a first-level child of the stream that is not supported by the server.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams unsupported-version": (
        "",
        "The value of the        'version' attribute provided by the initiating entity in the stream header specifies a version of XMPP that is not supported by the server.",
    ),
    "urn:ietf:params:xml:ns:xmpp-streams xml-not-well-formed": (
        "",
        "The initiating entity has sent XML that is not well-formed.",
    ),
}

_localpart_escape_chars = " \"&'/:<>@"


STREAM_NOT_AUTHORIZED = "urn:ietf:params:xml:ns:xmpp-streams not-authorized"
STREAM_REMOTE_CONNECTION_FAILED = (
    "urn:ietf:params:xml:ns:xmpp-streams remote-connection-failed"
)
SASL_MECHANISM_TOO_WEAK = "urn:ietf:params:xml:ns:xmpp-sasl mechanism-too-weak"
STREAM_XML_NOT_WELL_FORMED = "urn:ietf:params:xml:ns:xmpp-streams xml-not-well-formed"
ERR_JID_MALFORMED = "urn:ietf:params:xml:ns:xmpp-stanzas jid-malformed"
STREAM_SEE_OTHER_HOST = "urn:ietf:params:xml:ns:xmpp-streams see-other-host"
STREAM_BAD_NAMESPACE_PREFIX = "urn:ietf:params:xml:ns:xmpp-streams bad-namespace-prefix"
ERR_SERVICE_UNAVAILABLE = "urn:ietf:params:xml:ns:xmpp-stanzas service-unavailable"
STREAM_CONNECTION_TIMEOUT = "urn:ietf:params:xml:ns:xmpp-streams connection-timeout"
STREAM_UNSUPPORTED_VERSION = "urn:ietf:params:xml:ns:xmpp-streams unsupported-version"
STREAM_IMPROPER_ADDRESSING = "urn:ietf:params:xml:ns:xmpp-streams improper-addressing"
STREAM_UNDEFINED_CONDITION = "urn:ietf:params:xml:ns:xmpp-streams undefined-condition"
SASL_NOT_AUTHORIZED = "urn:ietf:params:xml:ns:xmpp-sasl not-authorized"
ERR_GONE = "urn:ietf:params:xml:ns:xmpp-stanzas gone"
SASL_TEMPORARY_AUTH_FAILURE = "urn:ietf:params:xml:ns:xmpp-sasl temporary-auth-failure"
ERR_REMOTE_SERVER_NOT_FOUND = (
    "urn:ietf:params:xml:ns:xmpp-stanzas remote-server-not-found"
)
ERR_UNEXPECTED_REQUEST = "urn:ietf:params:xml:ns:xmpp-stanzas unexpected-request"
ERR_RECIPIENT_UNAVAILABLE = "urn:ietf:params:xml:ns:xmpp-stanzas recipient-unavailable"
ERR_CONFLICT = "urn:ietf:params:xml:ns:xmpp-stanzas conflict"
STREAM_SYSTEM_SHUTDOWN = "urn:ietf:params:xml:ns:xmpp-streams system-shutdown"
STREAM_BAD_FORMAT = "urn:ietf:params:xml:ns:xmpp-streams bad-format"
ERR_SUBSCRIPTION_REQUIRED = "urn:ietf:params:xml:ns:xmpp-stanzas subscription-required"
STREAM_INTERNAL_SERVER_ERROR = (
    "urn:ietf:params:xml:ns:xmpp-streams internal-server-error"
)
ERR_NOT_AUTHORIZED = "urn:ietf:params:xml:ns:xmpp-stanzas not-authorized"
SASL_ABORTED = "urn:ietf:params:xml:ns:xmpp-sasl aborted"
ERR_REGISTRATION_REQUIRED = "urn:ietf:params:xml:ns:xmpp-stanzas registration-required"
ERR_INTERNAL_SERVER_ERROR = "urn:ietf:params:xml:ns:xmpp-stanzas internal-server-error"
SASL_INCORRECT_ENCODING = "urn:ietf:params:xml:ns:xmpp-sasl incorrect-encoding"
STREAM_HOST_GONE = "urn:ietf:params:xml:ns:xmpp-streams host-gone"
STREAM_POLICY_VIOLATION = "urn:ietf:params:xml:ns:xmpp-streams policy-violation"
STREAM_INVALID_XML = "urn:ietf:params:xml:ns:xmpp-streams invalid-xml"
STREAM_CONFLICT = "urn:ietf:params:xml:ns:xmpp-streams conflict"
STREAM_RESOURCE_CONSTRAINT = "urn:ietf:params:xml:ns:xmpp-streams resource-constraint"
STREAM_UNSUPPORTED_ENCODING = "urn:ietf:params:xml:ns:xmpp-streams unsupported-encoding"
ERR_NOT_ALLOWED = "urn:ietf:params:xml:ns:xmpp-stanzas not-allowed"
ERR_ITEM_NOT_FOUND = "urn:ietf:params:xml:ns:xmpp-stanzas item-not-found"
ERR_NOT_ACCEPTABLE = "urn:ietf:params:xml:ns:xmpp-stanzas not-acceptable"
STREAM_INVALID_FROM = "urn:ietf:params:xml:ns:xmpp-streams invalid-from"
ERR_FEATURE_NOT_IMPLEMENTED = (
    "urn:ietf:params:xml:ns:xmpp-stanzas feature-not-implemented"
)
ERR_BAD_REQUEST = "urn:ietf:params:xml:ns:xmpp-stanzas bad-request"
STREAM_INVALID_ID = "urn:ietf:params:xml:ns:xmpp-streams invalid-id"
STREAM_HOST_UNKNOWN = "urn:ietf:params:xml:ns:xmpp-streams host-unknown"
ERR_UNDEFINED_CONDITION = "urn:ietf:params:xml:ns:xmpp-stanzas undefined-condition"
SASL_INVALID_MECHANISM = "urn:ietf:params:xml:ns:xmpp-sasl invalid-mechanism"
STREAM_RESTRICTED_XML = "urn:ietf:params:xml:ns:xmpp-streams restricted-xml"
ERR_RESOURCE_CONSTRAINT = "urn:ietf:params:xml:ns:xmpp-stanzas resource-constraint"
ERR_REMOTE_SERVER_TIMEOUT = "urn:ietf:params:xml:ns:xmpp-stanzas remote-server-timeout"
SASL_INVALID_AUTHZID = "urn:ietf:params:xml:ns:xmpp-sasl invalid-authzid"
ERR_PAYMENT_REQUIRED = "urn:ietf:params:xml:ns:xmpp-stanzas payment-required"
STREAM_INVALID_NAMESPACE = "urn:ietf:params:xml:ns:xmpp-streams invalid-namespace"
ERR_REDIRECT = "urn:ietf:params:xml:ns:xmpp-stanzas redirect"
STREAM_UNSUPPORTED_STANZA_TYPE = (
    "urn:ietf:params:xml:ns:xmpp-streams unsupported-stanza-type"
)
ERR_FORBIDDEN = "urn:ietf:params:xml:ns:xmpp-stanzas forbidden"


def isResultNode(node: Protocol) -> bool:
    """
    Return true if the node is a positive reply
    """
    return node and node.getType() == "result"


def isErrorNode(node: Protocol) -> bool:
    """
    Return true if the node is a negative reply
    """
    return node and node.getType() == "error"


def isMucPM(message: Message) -> bool:
    muc_user = message.getTag("x", namespace=Namespace.MUC_USER)
    return (
        message.getType() in ("chat", "error")
        and muc_user is not None
        and not muc_user.getChildren()
    )


class NodeProcessed(Exception):
    """
    Exception that should be raised by handler when the handling should be
    stopped
    """


class StreamError(Exception):
    """
    Base exception class for stream errors
    """


class BadFormat(StreamError):
    pass


class BadNamespacePrefix(StreamError):
    pass


class Conflict(StreamError):
    pass


class ConnectionTimeout(StreamError):
    pass


class HostGone(StreamError):
    pass


class HostUnknown(StreamError):
    pass


class ImproperAddressing(StreamError):
    pass


class InternalServerError(StreamError):
    pass


class InvalidFrom(StreamError):
    pass


class InvalidID(StreamError):
    pass


class InvalidNamespace(StreamError):
    pass


class InvalidXML(StreamError):
    pass


class NotAuthorized(StreamError):
    pass


class PolicyViolation(StreamError):
    pass


class RemoteConnectionFailed(StreamError):
    pass


class ResourceConstraint(StreamError):
    pass


class RestrictedXML(StreamError):
    pass


class SeeOtherHost(StreamError):
    pass


class SystemShutdown(StreamError):
    pass


class UndefinedCondition(StreamError):
    pass


class UnsupportedEncoding(StreamError):
    pass


class UnsupportedStanzaType(StreamError):
    pass


class UnsupportedVersion(StreamError):
    pass


class XMLNotWellFormed(StreamError):
    pass


class InvalidStanza(Exception):
    pass


class InvalidJid(Exception):
    pass


class LocalpartByteLimit(InvalidJid):
    def __init__(self):
        InvalidJid.__init__(self, "Localpart must be between 1 and 1023 bytes")


class LocalpartNotAllowedChar(InvalidJid):
    def __init__(self):
        InvalidJid.__init__(self, "Not allowed character in localpart")


class ResourcepartByteLimit(InvalidJid):
    def __init__(self):
        InvalidJid.__init__(self, "Resourcepart must be between 1 and 1023 bytes")


class ResourcepartNotAllowedChar(InvalidJid):
    def __init__(self):
        InvalidJid.__init__(self, "Not allowed character in resourcepart")


class DomainpartByteLimit(InvalidJid):
    def __init__(self):
        InvalidJid.__init__(self, "Domainpart must be between 1 and 1023 bytes")


class DomainpartNotAllowedChar(InvalidJid):
    def __init__(self):
        InvalidJid.__init__(self, "Not allowed character in domainpart")


class StanzaMalformed(Exception):
    pass


class DiscoInfoMalformed(Exception):
    pass


stream_exceptions = {
    "bad-format": BadFormat,
    "bad-namespace-prefix": BadNamespacePrefix,
    "conflict": Conflict,
    "connection-timeout": ConnectionTimeout,
    "host-gone": HostGone,
    "host-unknown": HostUnknown,
    "improper-addressing": ImproperAddressing,
    "internal-server-error": InternalServerError,
    "invalid-from": InvalidFrom,
    "invalid-id": InvalidID,
    "invalid-namespace": InvalidNamespace,
    "invalid-xml": InvalidXML,
    "not-authorized": NotAuthorized,
    "policy-violation": PolicyViolation,
    "remote-connection-failed": RemoteConnectionFailed,
    "resource-constraint": ResourceConstraint,
    "restricted-xml": RestrictedXML,
    "see-other-host": SeeOtherHost,
    "system-shutdown": SystemShutdown,
    "undefined-condition": UndefinedCondition,
    "unsupported-encoding": UnsupportedEncoding,
    "unsupported-stanza-type": UnsupportedStanzaType,
    "unsupported-version": UnsupportedVersion,
    "xml-not-well-formed": XMLNotWellFormed,
}


def deprecation_warning(message: str) -> None:
    warnings.warn(message, DeprecationWarning, stacklevel=1)


def split_jid_string(jid_string: str) -> tuple[str | None, str, str | None]:

    # https://tools.ietf.org/html/rfc7622#section-3.2

    # Remove any portion from the first '/' character to the end of the
    # string (if there is a '/' character present).

    # Remove any portion from the beginning of the string to the first
    # '@' character (if there is an '@' character present).

    if jid_string.find("/") != -1:
        rest, resourcepart = jid_string.split("/", 1)
    else:
        rest, resourcepart = jid_string, None

    if rest.find("@") != -1:
        localpart, domainpart = rest.split("@", 1)
    else:
        localpart, domainpart = None, rest

    return localpart, domainpart, resourcepart


@functools.cache
def validate_localpart(localpart: str) -> str:
    if not localpart or len(localpart.encode()) > 1023:
        raise LocalpartByteLimit

    if os.environ.get("NBXMPP_ENFORCE_PRECIS") is None:
        try:
            return nodeprep(localpart)
        except Exception:
            try:
                return enforce_precis_username(localpart)
            except Exception:
                raise LocalpartNotAllowedChar

    try:
        return enforce_precis_username(localpart)
    except Exception:
        raise LocalpartNotAllowedChar


@functools.cache
def validate_resourcepart(resourcepart: str) -> str:
    if not resourcepart or len(resourcepart.encode()) > 1023:
        raise ResourcepartByteLimit

    if os.environ.get("NBXMPP_ENFORCE_PRECIS") is None:
        try:
            return resourceprep(resourcepart)
        except Exception:
            try:
                return enforce_precis_opaque(resourcepart)
            except Exception:
                raise ResourcepartNotAllowedChar

    try:
        return enforce_precis_opaque(resourcepart)
    except Exception:
        raise ResourcepartNotAllowedChar


@functools.cache
def validate_domainpart(domainpart: str) -> str:
    if not domainpart:
        raise DomainpartByteLimit

    ip_address = domainpart.strip("[]")
    if GLib.hostname_is_ip_address(ip_address):
        return ip_address

    length = len(domainpart.encode())
    if length == 0 or length > 1023:
        raise DomainpartByteLimit

    if domainpart.endswith("."):  # RFC7622, 3.2
        domainpart = domainpart[:-1]

    try:
        domainpart = idna2008_prep(domainpart)
    except Exception:
        raise DomainpartNotAllowedChar

    return domainpart


@functools.cache
def idna2008_prep(domain: str, to_ascii: bool = False) -> str:
    """
    Prepare with UTS46 case mapping to stay compatibel with the IDNA2003
    mapping. Further try to encode the domain to catch illegal domains.
    Only return the case mapped domain because on the XMPP wire,UTF8 domains
    are fine.
    """
    domain = idna.uts46_remap(domain)
    encoded_domain = idna.encode(domain)
    if to_ascii:
        return encoded_domain.decode()
    return domain


@functools.cache
def escape_localpart(localpart: str) -> str:
    # https://xmpp.org/extensions/xep-0106.html#bizrules-algorithm
    #
    # If there are any instances of character sequences that correspond
    # to escapings of the disallowed characters
    # (e.g., the character sequence "\27") or the escaping character
    # (i.e., the character sequence "\5c") in the source address,
    # the leading backslash character MUST be escaped to the character
    # sequence "\5c"

    for char in "\\" + _localpart_escape_chars:
        seq = "\\{:02x}".format(ord(char))
        localpart = localpart.replace(seq, "\\5c{:02x}".format(ord(char)))

    # Escape all other chars
    for char in _localpart_escape_chars:
        localpart = localpart.replace(char, "\\{:02x}".format(ord(char)))

    return localpart


@functools.cache
def unescape_localpart(localpart: str) -> str:
    if localpart.startswith("\\20") or localpart.endswith("\\20"):
        # Escaped JIDs are not allowed to start or end with \20
        # so this localpart must be already unescaped
        return localpart

    for char in _localpart_escape_chars:
        seq = "\\{:02x}".format(ord(char))
        localpart = localpart.replace(seq, char)

    for char in _localpart_escape_chars + "\\":
        seq = "\\5c{:02x}".format(ord(char))
        localpart = localpart.replace(seq, "\\{:02x}".format(ord(char)))

    return localpart


@dataclass(frozen=True, kw_only=True)
class JID:
    localpart: str | None
    domain: str
    resource: str | None

    def __init__(
        self,
        localpart: str | None,
        domain: str,
        resource: str | None,
    ):

        if localpart is not None:
            localpart = validate_localpart(localpart)
        object.__setattr__(self, "localpart", localpart)

        domain = validate_domainpart(domain)
        object.__setattr__(self, "domain", domain)

        if resource is not None:
            resource = validate_resourcepart(resource)
        object.__setattr__(self, "resource", resource)

    @classmethod
    @functools.cache
    def from_string(cls, jid_string: str, force_bare: bool = False) -> JID:
        localpart, domainpart, resourcepart = split_jid_string(jid_string)

        if force_bare:
            resourcepart = None

        try:
            return cls(localpart=localpart, domain=domainpart, resource=resourcepart)
        except Exception as error:
            raise InvalidJid('Unable to parse "%s"' % jid_string) from error

    @classmethod
    @functools.cache
    def from_user_input(cls, user_input: str, escaped: bool = False) -> JID:
        # Use this if we want JIDs to be escaped according to XEP-0106
        # The standard JID parsing cannot be applied because user_input is
        # not a valid JID.

        # Only user_input which after escaping result in a bare JID can be
        # successfully parsed.

        # The assumpution is user_input is a bare JID so we start with an
        # rsplit on @ because we assume there is no resource, so the char @
        # in the localpart can later be escaped.

        if escaped:
            # for convenience
            return cls.from_string(user_input)

        if "@" in user_input:
            localpart, domainpart = user_input.rsplit("@", 1)
            if localpart.startswith(" ") or localpart.endswith(" "):
                raise LocalpartNotAllowedChar

            localpart = escape_localpart(localpart)

        else:
            localpart = None
            domainpart = user_input

        try:
            return cls(localpart=localpart, domain=domainpart, resource=None)
        except Exception as error:
            raise InvalidJid('Unable to parse "%s"' % user_input) from error

    @classmethod
    @functools.cache
    def from_iri(cls, iri_str: str, *, force_bare: bool = False) -> JID:
        try:
            iri_str = clean_iri(iri_str)
        except ValueError as error:
            raise InvalidJid('Unable to parse "%s"' % iri_str) from error

        localpart, domainpart, resourcepart = split_jid_string(iri_str)

        if localpart is not None:
            localpart = GLib.Uri.unescape_string(localpart)

        if force_bare:
            resourcepart = None

        if resourcepart is not None:
            resourcepart = GLib.Uri.unescape_string(resourcepart)

        try:
            return cls(localpart=localpart, domain=domainpart, resource=resourcepart)
        except Exception as error:
            raise InvalidJid('Unable to parse "%s"' % iri_str) from error

    def __str__(self) -> str:
        if self.localpart:
            jid = f"{self.localpart}@{self.domain}"
        else:
            jid = self.domain

        if self.resource is not None:
            return f"{jid}/{self.resource}"
        return jid

    def __conform__(self, protocol: sqlite3.PrepareProtocol | Any) -> str:
        if protocol is sqlite3.PrepareProtocol:
            return str(self)
        raise ValueError

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            try:
                return JID.from_string(other) == self
            except Exception:
                return False

        if isinstance(other, JID):
            return (
                self.localpart == other.localpart
                and self.domain == other.domain
                and self.resource == other.resource
            )

        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def domain_to_ascii(self) -> str:
        return idna2008_prep(self.domain, to_ascii=True)

    @property
    def bare(self) -> str:
        if self.localpart is not None:
            return f"{self.localpart}@{self.domain}"
        return self.domain

    @property
    def is_bare(self) -> bool:
        return self.resource is None

    def new_as_bare(self) -> JID:
        if self.resource is None:
            return self
        return JID.from_string(self.bare)

    def bare_match(self, other: str | JID) -> bool:
        if isinstance(other, str):
            other = JID.from_string(other)
        return self.bare == other.bare

    @property
    def is_domain(self) -> bool:
        return self.localpart is None and self.resource is None

    @property
    def is_full(self) -> bool:
        return self.localpart is not None and self.resource is not None

    def new_with(self, **kwargs: Any) -> JID:
        new = asdict(self)
        new.update(kwargs)
        return JID(**new)

    def to_user_string(self, show_punycode: bool = True) -> str:
        domain = self.domain_to_ascii()
        if domain.startswith("xn--") and show_punycode:
            domain_encoded = f" ({domain})"
        else:
            domain_encoded = ""

        if self.localpart is None:
            return f"{self}{domain_encoded}"

        localpart = unescape_localpart(self.localpart)

        if self.resource is None:
            return f"{localpart}@{self.domain}{domain_encoded}"
        return f"{localpart}@{self.domain}/{self.resource}{domain_encoded}"

    def to_iri(
        self,
        query: str | tuple[str, list[tuple[str, str]]] | None = None,
        fragment: str | None = None,
    ) -> str:

        if self.localpart:
            inode = escape_inode(self.localpart)
            ipathxmpp = f"{inode}@{self.domain}"
        else:
            ipathxmpp = f"{self.domain}"

        if self.resource is not None:
            ires = escape_ires(self.resource)
            ipathxmpp = f"{ipathxmpp}/{ires}"

        iri = f"xmpp:{ipathxmpp}"

        if query is not None:
            if isinstance(query, str):
                querytype = query
                queryparams = None
            else:
                querytype, queryparams = query

            iquerytype = validate_querytype(querytype)
            iri += f"?{iquerytype}"

            if queryparams is not None:
                for ikey, ivalue in queryparams:
                    ivalue = escape_ivalue(ivalue)
                    ikey = validate_ikey(ikey)
                    iri += f";{ikey}={ivalue}"

        if fragment is not None:
            ifragment = escape_ifragment(fragment)
            iri += f"#{ifragment}"

        return iri

    def copy(self) -> JID:
        deprecation_warning("copy() is not needed, JID is immutable")
        return self


class StreamErrorNode(Node):
    def __init__(self, node: Protocol) -> None:
        Node.__init__(self, node=node)

        self._text: dict[str | None, str] = {}

        text_elements = self.getTags("text", namespace=Namespace.XMPP_STREAMS)
        for element in text_elements:
            lang = element.getXmlLang()
            text = element.getData()
            self._text[lang] = text

    def get_condition(self) -> str | None:
        for tag in self.getChildren():
            if tag.getName() != "text" and tag.getNamespace() == Namespace.XMPP_STREAMS:
                return tag.getName()
        return None

    def get_text(self, pref_lang: str | None = None) -> str:
        if pref_lang is not None:
            text = self._text.get(pref_lang)
            if text is not None:
                return text

        if self._text:
            text = self._text.get("en")
            if text is not None:
                return text

            text = self._text.get(None)
            if text is not None:
                return text
            return self._text.popitem()[1]
        return ""


class Protocol(Node):
    """
    A "stanza" object class. Contains methods that are common for presences, iqs
    and messages
    """

    def __init__(
        self,
        name: str | None = None,
        to: JID | str | None = None,
        typ: str | None = None,
        frm: JID | str | None = None,
        attrs: dict[str, Any] | None = None,
        payload: list[Node | str] | None = None,
        timestamp: str | None = None,
        xmlns: str | None = None,
        node: Node | bytes | str | None = None,
    ) -> None:
        """
        Constructor, name is the name of the stanza
        i.e. 'message' or 'presence'or 'iq'

        to is the value of 'to' attribure, 'typ' - 'type' attribute
        frn - from attribure, attrs - other attributes mapping,
        payload - same meaning as for simplexml payload definition
        timestamp - the time value that needs to be stamped over stanza
        xmlns - namespace of top stanza node
        node - parsed or unparsed stana to be taken as prototype.
        """
        if not attrs:
            attrs = {}
        if to:
            attrs["to"] = to
        if frm:
            attrs["from"] = frm
        if typ:
            attrs["type"] = typ
        Node.__init__(self, tag=name, attrs=attrs, payload=payload, node=node)
        if not node and xmlns:
            self.setNamespace(xmlns)
        if self["to"]:
            self.setTo(self["to"])
        if self["from"]:
            self.setFrom(self["from"])
        if (
            node
            and isinstance(node, Protocol)
            and self.__class__ == node.__class__
            and "id" in self.attrs
        ):
            del self.attrs["id"]
        self.timestamp: str | None = None
        for d in self.getTags("delay", namespace=Namespace.DELAY2):
            try:
                if d.getAttr("stamp") < self.getTimestamp2():
                    self.setTimestamp(d.getAttr("stamp"))
            except Exception:
                pass
        if not self.timestamp:
            for x in self.getTags("x", namespace=Namespace.DELAY):
                try:
                    if x.getAttr("stamp") < self.getTimestamp():
                        self.setTimestamp(x.getAttr("stamp"))
                except Exception:
                    pass
        if timestamp is not None:
            self.setTimestamp(timestamp)

    def isError(self) -> bool:
        return self.getAttr("type") == "error"

    def isResult(self) -> bool:
        return self.getAttr("type") == "result"

    def getTo(self):
        """
        Return value of the 'to' attribute
        """
        try:
            return self["to"]
        except Exception:
            pass
        return None

    def getFrom(self):
        """
        Return value of the 'from' attribute
        """
        try:
            return self["from"]
        except Exception:
            pass
        return None

    def getTimestamp(self) -> str:
        """
        Return the timestamp in the 'yyyymmddThhmmss' format
        """
        if self.timestamp:
            return self.timestamp
        return time.strftime("%Y%m%dT%H:%M:%S", time.gmtime())

    def getTimestamp2(self) -> str:
        """
        Return the timestamp in the 'yyyymmddThhmmss' format
        """
        if self.timestamp:
            return self.timestamp
        return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())

    def getJid(self):
        """
        Return the value of the 'jid' attribute
        """
        attr = self.getAttr("jid")
        if attr:
            return JID.from_string(attr)
        return attr

    def getID(self) -> str | None:
        """
        Return the value of the 'id' attribute
        """
        return self.getAttr("id")

    def setTo(self, val: str | JID):
        """
        Set the value of the 'to' attribute
        """
        if not isinstance(val, JID):
            val = JID.from_string(val)
        self.setAttr("to", val)

    def getType(self) -> str | None:
        """
        Return the value of the 'type' attribute
        """
        return self.getAttr("type")

    def setFrom(self, val: str | JID) -> None:
        """
        Set the value of the 'from' attribute
        """
        if not isinstance(val, JID):
            val = JID.from_string(val)
        self.setAttr("from", val)

    def setType(self, val: str) -> None:
        """
        Set the value of the 'type' attribute
        """
        self.setAttr("type", val)

    def setID(self, val: str) -> None:
        """
        Set the value of the 'id' attribute
        """
        self.setAttr("id", val)

    def getError(self) -> str | None:
        """
        Return the error-condition (if present) or the textual description
        of the error (otherwise)
        """
        errtag = self.getTag("error")
        if errtag is None:
            return None
        for tag in errtag.getChildren():
            if tag.getName() != "text" and tag.getNamespace() == Namespace.STANZAS:
                return tag.getName()
        return None

    def getAppError(self) -> str | None:
        errtag = self.getTag("error")
        if errtag is None:
            return None
        for tag in errtag.getChildren():
            if tag.getName() != "text" and tag.getNamespace() != Namespace.STANZAS:
                return tag.getName()
        return None

    def getAppErrorNamespace(self) -> str | None:
        errtag = self.getTag("error")
        if errtag is None:
            return None
        for tag in errtag.getChildren():
            if tag.getName() != "text" and tag.getNamespace() != Namespace.STANZAS:
                return tag.getNamespace()
        return None

    def getErrorMsg(self) -> str | None:
        """
        Return the textual description of the error (if present)
        or the error condition
        """
        errtag = self.getTag("error")
        if errtag:
            for tag in errtag.getChildren():
                if tag.getName() == "text":
                    return tag.getData()
            return self.getError()
        return None

    def getErrorType(self) -> str | None:
        """
        Return the error type.
        """
        return self.getTagAttr("error", "type")

    def getStatusConditions(self, as_code: bool = False) -> list[str]:
        """
        Return the status conditions list as defined in XEP-0306.
        """
        result: set[str] = set()
        status_tags = self.getTags("status")
        for status in status_tags:
            if as_code:
                code = status.getAttr("code")
                if code is not None:
                    result.add(code)
            else:
                for condition in status.getChildren():
                    result.add(condition.getName())
        return list(result)

    def setError(self, error: ErrorNode | str) -> None:
        """
        Set the error-conditions
        """
        if isinstance(error, str):
            error = ErrorNode(error)
        self.setType("error")
        self.addChild(node=error)

    def setTimestamp(self, val: str | None = None) -> None:
        """
        Set the timestamp. timestamp should be the yyyymmddThhmmss string
        """
        if not val:
            val = time.strftime("%Y%m%dT%H:%M:%S", time.gmtime())
        self.timestamp = val
        self.setTag("x", {"stamp": self.timestamp}, namespace=Namespace.DELAY)

    def getProperties(self) -> list[str]:
        """
        Return the list of namespaces to which belongs the
        direct childs of element
        """
        props: list[str] = []
        for child in self.getChildren():
            prop = child.getNamespace()
            if prop not in props:
                props.append(prop)
        return props

    def getTag(
        self,
        name: str,
        attrs: dict[str, Any] | None = None,
        namespace: str | None = None,
        protocol: bool = False,
    ) -> Node | None:
        """
        Return the Node instance for the tag.
        If protocol is True convert to a new Protocol/Message instance.
        """
        tag = Node.getTag(self, name, attrs, namespace)
        if protocol and tag:
            if name == "message":
                return Message(node=tag)
            return Protocol(node=tag)
        return tag

    def __setitem__(self, item: str, val: str | JID) -> None:
        """
        Set the item 'item' to the value 'val'
        """
        if item in ["to", "from"]:
            if not isinstance(val, JID):
                val = JID.from_string(val)
        self.setAttr(item, val)


class Message(Protocol):
    """
    XMPP Message stanza - "push" mechanism
    """

    def __init__(
        self,
        to: str | None = None,
        body: str | None = None,
        xhtml: str | None = None,
        typ: Literal["chat"] | Literal["groupchat"] | Literal["error"] | None = None,
        subject: MucSubject | None = None,
        attrs: dict[str, Any] | None = None,
        frm: JID | str | None = None,
        payload: list[Node | str] | None = None,
        timestamp: str | None = None,
        xmlns: str = Namespace.CLIENT,
        node: Node | str | None = None,
    ) -> None:
        """
        You can specify recipient, text of message, type of message any
        additional attributes, sender of the message, any additional payload
        (f.e. jabber:x:delay element) and namespace in one go.

        Alternatively you can pass in the other XML object as the 'node'
        parameted to replicate it as message
        """
        Protocol.__init__(
            self,
            "message",
            to=to,
            typ=typ,
            attrs=attrs,
            frm=frm,
            payload=payload,
            timestamp=timestamp,
            xmlns=xmlns,
            node=node,
        )
        if body:
            self.setBody(body)
        if xhtml is not None:
            self.setXHTML(xhtml)
        if subject is not None:
            self.setSubject(subject)

    def getBody(self) -> str | None:
        """
        Return text of the message
        """
        return self.getTagData("body")

    def getXHTML(self) -> Node | None:
        return self.getTag("html", namespace=Namespace.XHTML_IM)

    def getSubject(self) -> str | None:
        """
        Return subject of the message
        """
        return self.getTagData("subject")

    def getThread(self) -> str | None:
        """
        Return thread of the message
        """
        return self.getTagData("thread")

    def getOriginID(self) -> str | None:
        """
        Return origin-id of the message
        """
        return self.getTagAttr("origin-id", namespace=Namespace.SID, attr="id")

    def getStanzaIDAttrs(self) -> tuple[str | None, str | None]:
        """
        Return the stanza-id attributes of the message
        """
        try:
            attrs = self.getTag("stanza-id", namespace=Namespace.SID).getAttrs()
        except Exception:
            return None, None
        return attrs["id"], attrs["by"]

    def setBody(self, val: str) -> None:
        """
        Set the text of the message"""
        self.setTagData("body", val)

    def setXHTML(self, body: str | Node, add: bool = False) -> None:
        if isinstance(body, str):
            body = Node(node=body)
        if add:
            xhtml = self.getTag("html", namespace=Namespace.XHTML_IM)
            if xhtml is not None:
                xhtml.addChild(node=body)
            else:
                self.addChild("html", namespace=Namespace.XHTML_IM, payload=body)
        else:
            xhtml_nodes = self.getTags("html", namespace=Namespace.XHTML_IM)
            for xhtml in xhtml_nodes:
                self.delChild(xhtml)
            self.addChild("html", namespace=Namespace.XHTML_IM, payload=body)

    def setSubject(self, val: MucSubject | None) -> None:
        """
        Set the subject of the message
        """
        self.setTagData("subject", val)

    def setThread(self, val: str) -> None:
        """
        Set the thread of the message
        """
        self.setTagData("thread", val)

    def setOriginID(self, val: str) -> None:
        """
        Sets the origin-id of the message
        """
        self.setTag("origin-id", namespace=Namespace.SID, attrs={"id": val})

    def buildReply(self, text: str | None = None) -> Message:
        """
        Builds and returns another message object with specified text. The to,
        from, thread and type properties of new message are pre-set as reply to
        this message
        """
        m = Message(to=self.getFrom(), frm=self.getTo(), body=text, typ=self.getType())
        th = self.getThread()
        if th:
            m.setThread(th)
        return m

    def getStatusCode(self) -> list[str | None]:
        """
        Return the status code of the message (for groupchat config change)
        """
        attrs: list[str | None] = []
        for xtag in self.getTags("x"):
            for child in xtag.getTags("status"):
                attrs.append(child.getAttr("code"))
        return attrs

    def setMarker(self, type_: str, id_: str) -> None:
        self.setTag(type_, namespace=Namespace.CHATMARKERS, attrs={"id": id_})

    def setMarkable(self) -> None:
        self.setTag("markable", namespace=Namespace.CHATMARKERS)

    def setReceiptRequest(self) -> None:
        self.setTag("request", namespace=Namespace.RECEIPTS)

    def setReceiptReceived(self, id_: str) -> None:
        self.setTag("received", namespace=Namespace.RECEIPTS, attrs={"id": id_})

    def setReply(
        self,
        recipient_jid: str,
        reply_to_id: str,
        fallback_start: int,
        fallback_end: int,
    ) -> None:

        self.setTag(
            "reply",
            namespace=Namespace.REPLY,
            attrs={"id": reply_to_id, "to": recipient_jid},
        )

        fallback_tag = self.setTag(
            "fallback", namespace=Namespace.FALLBACK, attrs={"for": Namespace.REPLY}
        )
        fallback_tag.addChild(
            "body", attrs={"start": str(fallback_start), "end": str(fallback_end)}
        )

    def setOOB(self, url: str, desc: str | None = None) -> None:
        oob = self.setTag("x", namespace=Namespace.X_OOB)
        oob.setTagData("url", url)
        if desc is not None:
            oob.setTagData("desc", desc)

    def setCorrection(self, id_: str) -> None:
        self.setTag("replace", namespace=Namespace.CORRECT, attrs={"id": id_})

    def setAttention(self) -> None:
        self.setTag("attention", namespace=Namespace.ATTENTION)

    def setHint(self, hint: str) -> None:
        self.setTag(hint, namespace=Namespace.HINTS)

    def setReactions(self, target_id: str, emojis: Iterable[str]) -> None:
        reactions = self.addChild(
            "reactions", namespace=Namespace.REACTIONS, attrs={"id": target_id}
        )
        for e in emojis:
            reactions.addChild("reaction", namespace=Namespace.REACTIONS, payload=[e])

    def getReactions(self) -> tuple[str, set[str]] | None:
        reactions = self.getTag("reactions", namespace=Namespace.REACTIONS)
        if not reactions:
            return None

        react_to = reactions.getAttr("id")
        if not react_to:
            return None

        tags = reactions.getTags("reaction", namespace=Namespace.REACTIONS)

        # strip() in case clients surround emojis with whitespace
        return react_to, {t.getData().strip() for t in tags}

    def setRetracted(self, target_id: str, fallback_text: str | None = None) -> None:
        retract = self.addChild("retract", namespace=Namespace.MESSAGE_RETRACT_1)
        retract.setAttr("id", target_id)

        if not fallback_text:
            return

        self.setTag(
            "fallback",
            namespace=Namespace.FALLBACK,
            attrs={"for": Namespace.MESSAGE_RETRACT_1},
        )

        self.setBody(fallback_text)


class Presence(Protocol):

    def __init__(
        self,
        to: JID | str | None = None,
        typ: Literal["probe", "unavailable"] | None = None,
        priority: int | None = None,
        show: Literal["chat", "away", "xa", "dnd"] | None = None,
        status: str | None = None,
        attrs: dict[str, Any] | None = None,
        frm: JID | None = None,
        timestamp: str | None = None,
        payload: list[Node | str] | None = None,
        xmlns: str = Namespace.CLIENT,
        node: Node | str | None = None,
    ) -> None:
        """
        You can specify recipient, type of message, priority, show and status
        values any additional attributes, sender of the presence, timestamp, any
        additional payload (f.e. jabber:x:delay element) and namespace in one
        go. Alternatively you can pass in the other XML object as the 'node'
        parameted to replicate it as presence
        """
        Protocol.__init__(
            self,
            "presence",
            to=to,
            typ=typ,
            attrs=attrs,
            frm=frm,
            payload=payload,
            timestamp=timestamp,
            xmlns=xmlns,
            node=node,
        )
        if priority:
            self.setPriority(priority)
        if show:
            self.setShow(show)
        if status:
            self.setStatus(status)

    def getPriority(self) -> int:
        """
        Return the priority of the message
        """
        return self.getTagData("priority")

    def getShow(self) -> str:
        """
        Return the show value of the message
        """
        return self.getTagData("show")

    def getStatus(self) -> str:
        """
        Return the status string of the message
        """
        return self.getTagData("status") or ""

    def setPriority(self, val: int | None) -> None:
        """
        Set the priority of the message
        """
        self.setTagData("priority", val)

    def setShow(self, val: str) -> None:
        """
        Set the show value of the message
        """
        if val not in ["away", "chat", "dnd", "xa"]:
            raise ValueError("Invalid show value: %s" % val)
        self.setTagData("show", val)

    def setStatus(self, val: str) -> None:
        """
        Set the status string of the message
        """
        self.setTagData("status", val)

    def _muc_getItemAttr(self, tag: str, attr: str):
        for xtag in self.getTags("x"):
            if xtag.getNamespace() not in (Namespace.MUC_USER, Namespace.MUC_ADMIN):
                continue
            for child in xtag.getTags(tag):
                return child.getAttr(attr)

    def _muc_getSubTagDataAttr(self, tag: str, attr: str):
        for xtag in self.getTags("x"):
            if xtag.getNamespace() not in (Namespace.MUC_USER, Namespace.MUC_ADMIN):
                continue
            for child in xtag.getTags("item"):
                for cchild in child.getTags(tag):
                    return cchild.getData(), cchild.getAttr(attr)
        return None, None

    def getRole(self):
        """
        Return the presence role (for groupchat)
        """
        return self._muc_getItemAttr("item", "role")

    def getAffiliation(self):
        """
        Return the presence affiliation (for groupchat)
        """
        return self._muc_getItemAttr("item", "affiliation")

    def getNewNick(self):
        """
        Return the status code of the presence (for groupchat)
        """
        return self._muc_getItemAttr("item", "nick")

    def getJid(self):
        """
        Return the presence jid (for groupchat)
        """
        return self._muc_getItemAttr("item", "jid")

    def getReason(self):
        """
        Returns the reason of the presence (for groupchat)
        """
        return self._muc_getSubTagDataAttr("reason", "")[0]

    def getActor(self):
        """
        Return the reason of the presence (for groupchat)
        """
        return self._muc_getSubTagDataAttr("actor", "jid")[1]

    def getStatusCode(self):
        """
        Return the status code of the presence (for groupchat)
        """
        attrs = []
        for xtag in self.getTags("x"):
            for child in xtag.getTags("status"):
                attrs.append(child.getAttr("code"))
        return attrs


class Iq(Protocol):
    """
    XMPP Iq object - get/set dialog mechanism
    """

    def __init__(
        self,
        typ: str | None = None,
        queryNS: str | None = None,  # noqa: N803
        attrs: dict[str, str] | None = None,
        to: JID | str | None = None,
        frm: JID | str | None = None,
        payload: list[Node | str] | Node | None = None,
        xmlns: str = Namespace.CLIENT,
        node: Node | None = None,
    ) -> None:
        """
        You can specify type, query namespace any additional attributes,
        recipient of the iq, sender of the iq, any additional payload (f.e.
        jabber:x:data node) and namespace in one go.

        Alternatively you can pass in the other XML object as the 'node'
        parameted to replicate it as an iq
        """
        Protocol.__init__(
            self, "iq", to=to, typ=typ, attrs=attrs, frm=frm, xmlns=xmlns, node=node
        )
        if payload:
            self.setQueryPayload(payload)
        if queryNS:
            self.setQueryNS(queryNS)

    def getQuery(self) -> Node | None:
        """
        Return the IQ's child element if it exists, None otherwise.
        """
        children = self.getChildren()
        if children and self.getType() != "error" and children[0].getName() != "error":
            return children[0]
        return None

    def getQueryNS(self) -> str | None:
        """
        Return the namespace of the 'query' child element
        """
        tag = self.getQuery()
        if tag:
            return tag.getNamespace()
        return None

    def getQuerynode(self) -> str | None:
        """
        Return the 'node' attribute value of the 'query' child element
        """
        tag = self.getQuery()
        if tag:
            return tag.getAttr("node")
        return None

    def getQueryChildren(self) -> list[Node] | None:
        """
        Return the 'query' child element child nodes
        """
        tag = self.getQuery()
        if tag:
            return tag.getChildren()
        return None

    def getQueryChild(self, name: str | None = None) -> Node | None:
        """
        Return the 'query' child element with name, or the first element
        which is not an error element
        """
        query = self.getQuery()
        if not query:
            return None
        for node in query.getChildren():
            if name is not None:
                if node.getName() == name:
                    return node
            else:
                if node.getName() != "error":
                    return node
        return None

    def setQuery(self, name: str | None = None) -> Node:
        """
        Change the name of the query node, creating it if needed. Keep the
        existing name if none is given (use 'query' if it's a creation).
        Return the query node.
        """
        query = self.getQuery()
        if query is None:
            query = self.addChild("query")
        if name is not None:
            query.setName(name)
        return query

    def setQueryNS(self, namespace: str) -> None:
        """
        Set the namespace of the 'query' child element
        """
        self.setQuery().setNamespace(namespace)

    def setQueryPayload(self, payload: list[Node | str]):
        """
        Set the 'query' child element payload
        """
        self.setQuery().setPayload(payload)

    def setQuerynode(self, node: str) -> None:
        """
        Set the 'node' attribute value of the 'query' child element
        """
        self.setQuery().setAttr("node", node)

    def buildReply(self, typ: str) -> Iq:
        """
        Build and return another Iq object of specified type. The to, from and
        query child node of new Iq are pre-set as reply to this Iq.
        """
        iq = Iq(typ, to=self.getFrom(), frm=self.getTo(), attrs={"id": self.getID()})
        iq.setQuery(self.getQuery().getName()).setNamespace(self.getQueryNS())
        return iq

    def buildSimpleReply(self, typ: str) -> Iq:
        return Iq(typ, to=self.getFrom(), attrs={"id": self.getID()})


class Hashes(Node):
    """
    Hash elements for various XEPs as defined in XEP-300

    RECOMENDED HASH USE:
    Algorithm     Support
    MD2           MUST NOT
    MD4           MUST NOT
    MD5           MAY
    SHA-1         MUST
    SHA-256       MUST
    SHA-512       SHOULD
    """

    supported = ("md5", "sha-1", "sha-256", "sha-512")

    def __init__(self, nsp: str = Namespace.HASHES) -> None:
        Node.__init__(self, None, {}, [], None, None, False, None)
        self.setNamespace(nsp)
        self.setName("hash")

    def calculateHash(self, algo: str, file_string: str | bytes) -> str | None:
        """
        Calculate the hash and add it. It is preferable doing it here
        instead of doing it all over the place in Gajim.
        """
        hl = None
        hash_ = None
        # file_string can be a string or a file
        if isinstance(file_string, str):
            if algo == "sha-1":
                hl = hashlib.sha1()
            elif algo == "md5":
                hl = hashlib.md5()
            elif algo == "sha-256":
                hl = hashlib.sha256()
            elif algo == "sha-512":
                hl = hashlib.sha512()
            if hl:
                hl.update(file_string)
                hash_ = hl.hexdigest()
        else:  # if it is a file
            if algo == "sha-1":
                hl = hashlib.sha1()
            elif algo == "md5":
                hl = hashlib.md5()
            elif algo == "sha-256":
                hl = hashlib.sha256()
            elif algo == "sha-512":
                hl = hashlib.sha512()
            if hl:
                for line in file_string:
                    hl.update(line)
                hash_ = hl.hexdigest()
        return hash_

    def addHash(self, hash_: str, algo: str) -> None:
        self.setAttr("algo", algo)
        self.setData(hash_)


class Hashes2(Node):
    """
    Hash elements for various XEPs as defined in XEP-300

    RECOMENDED HASH USE:
    Algorithm     Support
    MD2           MUST NOT
    MD4           MUST NOT
    MD5           MUST NOT
    SHA-1         SHOULD NOT
    SHA-256       MUST
    SHA-512       SHOULD
    SHA3-256      MUST
    SHA3-512      SHOULD
    BLAKE2b256    MUST
    BLAKE2b512    SHOULD
    """

    supported = (
        "sha-256",
        "sha-512",
        "sha3-256",
        "sha3-512",
        "blake2b-256",
        "blake2b-512",
    )

    def __init__(self, nsp: str = Namespace.HASHES_2) -> None:
        Node.__init__(self, None, {}, [], None, None, False, None)
        self.setNamespace(nsp)
        self.setName("hash")

    def calculateHash(self, algo: str, file_string: str | bytes) -> str | None:
        """
        Calculate the hash and add it. It is preferable doing it here
        instead of doing it all over the place in Gajim.
        """
        hl = None
        hash_ = None
        if algo == "sha-256":
            hl = hashlib.sha256()
        elif algo == "sha-512":
            hl = hashlib.sha512()
        elif algo == "sha3-256":
            hl = hashlib.sha3_256()
        elif algo == "sha3-512":
            hl = hashlib.sha3_512()
        elif algo == "blake2b-256":
            hl = hashlib.blake2b(digest_size=32)
        elif algo == "blake2b-512":
            hl = hashlib.blake2b(digest_size=64)
        # file_string can be a string or a file
        if hl is not None:
            if isinstance(file_string, bytes):
                hl.update(file_string)
            else:  # if it is a file
                for line in file_string:
                    hl.update(line)
            hash_ = b64encode(hl.digest()).decode("ascii")
        return hash_

    def addHash(self, hash_: str, algo: str) -> None:
        self.setAttr("algo", algo)
        self.setData(hash_)


class BindRequest(Iq):
    def __init__(self, resource: str | None) -> None:
        if resource is None:
            res = resource
        else:
            res = Node("resource", payload=resource)
        Iq.__init__(self, typ="set")
        self.addChild(node=Node("bind", {"xmlns": Namespace.BIND}, payload=res))


class TLSRequest(Node):
    def __init__(self) -> None:
        Node.__init__(self, tag="starttls", attrs={"xmlns": Namespace.TLS})


class SessionRequest(Iq):
    def __init__(self) -> None:
        Iq.__init__(self, typ="set")
        self.addChild(node=Node("session", attrs={"xmlns": Namespace.SESSION}))


class StreamHeader(Node):
    def __init__(self, domain: str, lang: str | None = None) -> None:
        if lang is None:
            lang = "en"
        Node.__init__(
            self,
            tag="stream:stream",
            attrs={
                "xmlns": Namespace.CLIENT,
                "version": "1.0",
                "xmlns:stream": Namespace.STREAMS,
                "to": domain,
                "xml:lang": lang,
            },
        )


class WebsocketOpenHeader(Node):
    def __init__(self, domain: str, lang: str | None = None) -> None:
        if lang is None:
            lang = "en"
        Node.__init__(
            self,
            tag="open",
            attrs={
                "xmlns": Namespace.FRAMING,
                "version": "1.0",
                "to": domain,
                "xml:lang": lang,
            },
        )


class WebsocketCloseHeader(Node):
    def __init__(self) -> None:
        Node.__init__(self, tag="close", attrs={"xmlns": Namespace.FRAMING})


class Features(Node):
    def __init__(self, node: Node) -> None:
        Node.__init__(self, node=node)

    def has_starttls(self) -> tuple[bool, bool]:
        tls = self.getTag("starttls", namespace=Namespace.TLS)
        if tls is not None:
            required = tls.getTag("required") is not None
            return True, required
        return False, False

    def has_sasl(self) -> bool:
        return self.getTag("mechanisms", namespace=Namespace.SASL) is not None

    def has_sasl_2(self) -> bool:
        return self.getTag("authentication", namespace=Namespace.SASL2) is not None

    def get_mechs(self) -> set[str]:
        mechanisms = self.getTag("authentication", namespace=Namespace.SASL2)
        if mechanisms is None:
            mechanisms = self.getTag("mechanisms", namespace=Namespace.SASL)

        if mechanisms is None:
            return set()
        mechanisms = mechanisms.getTags("mechanism")
        return {mech.getData() for mech in mechanisms}

    def get_domain_based_name(self) -> str | None:
        hostname = self.getTag("hostname", namespace=Namespace.DOMAIN_BASED_NAME)
        if hostname is not None:
            return hostname.getData()
        return None

    def has_bind(self) -> bool:
        return self.getTag("bind", namespace=Namespace.BIND) is not None

    def session_required(self) -> bool:
        session = self.getTag("session", namespace=Namespace.SESSION)
        if session is not None:
            optional = session.getTag("optional") is not None
            return not optional
        return False

    def has_sm(self) -> bool:
        return self.getTag("sm", namespace=Namespace.STREAM_MGMT) is not None

    def has_roster_version(self) -> bool:
        return self.getTag("ver", namespace=Namespace.ROSTER_VER) is not None

    def has_register(self) -> bool:
        return self.getTag("register", namespace=Namespace.REGISTER_FEATURE) is not None

    def has_anonymous(self) -> bool:
        return "ANONYMOUS" in self.get_mechs()

    def get_channel_binding_type(self) -> Gio.TlsChannelBindingType | None:
        sasl_cb = self.getTag(
            "sasl-channel-binding", namespace=Namespace.CHANNEL_BINDING
        )
        if sasl_cb is None:
            return None

        exporter = sasl_cb.getTag("channel-binding", attrs={"type": "tls-exporter"})
        if exporter is not None:
            return Gio.TlsChannelBindingType.EXPORTER

        server_end_point = sasl_cb.getTag(
            "channel-binding", attrs={"type": "tls-server-end-point"}
        )
        if server_end_point is not None:
            return Gio.TlsChannelBindingType.SERVER_END_POINT
        return None


class ErrorNode(Node):
    """
    XMPP-style error element

    In the case of stanza error should be attached to XMPP stanza.
    In the case of stream-level errors should be used separately.
    """

    def __init__(
        self, name: str, typ: str | None = None, text: str | None = None
    ) -> None:
        """
        Mandatory parameter: name - name of error condition.
        Optional parameters: typ, text.
        Used for backwards compartibility with older jabber protocol.
        """
        if name in ERRORS:
            type_, txt = ERRORS[name]
            ns = name.split()[0]
        else:
            ns, type_, txt = Namespace.STANZAS, "cancel", ""
        if typ:
            type_ = typ
        if text:
            txt = text
        Node.__init__(self, "error", {}, [Node(name)])
        if type_:
            self.setAttr("type", type_)
        else:
            self.setName("stream:error")
        if txt:
            self.addChild(node=Node(ns + " text", {}, [txt]))


class Error(Protocol):
    """
    Used to quickly transform received stanza into error reply
    """

    def __init__(self, node: Protocol, error: ErrorNode | str, reply: int = 1) -> None:
        """
        Create error reply basing on the received 'node' stanza and the 'error'
        error condition

        If the 'node' is not the received stanza but locally created ('to' and
        'from' fields needs not swapping) specify the 'reply' argument as false.
        """
        if node.getType() == "error":
            raise ValueError("Can’t create error reply from error")

        if reply:
            Protocol.__init__(self, to=node.getFrom(), node=node)
            self.delAttr("from")
        else:
            Protocol.__init__(self, node=node)
        self.setError(error)


class DataField(Node):
    """
    This class is used in the DataForm class to describe the single data item

    If you are working with jabber:x:data (XEP-0004, XEP-0068, XEP-0122) then
    you will need to work with instances of this class.
    """

    def __init__(
        self,
        name: str | None = None,
        value=None,
        typ: str | None = None,
        required: int = 0,
        desc: str | None = None,
        options=None,
        node=None,
    ) -> None:
        """
        Create new data field of specified name,value and type

        Also 'required','desc' and 'options' fields can be set. Alternatively
        other XML object can be passed in as the 'node' parameted
        to replicate it as a new datafiled.
        """
        Node.__init__(self, "field", node=node)
        if name:
            self.setVar(name)
        if isinstance(value, list | tuple):
            self.setValues(value)
        elif value:
            self.setValue(value)
        if typ:
            self.setType(typ)
        elif not typ and not node:
            self.setType("text-single")
        if required:
            self.setRequired(required)
        if desc:
            self.setDesc(desc)
        if options:
            self.setOptions(options)

    def setRequired(self, req: int = 1) -> None:
        """
        Change the state of the 'required' flag
        """
        if req:
            self.setTag("required")
        else:
            try:
                self.delChild("required")
            except ValueError:
                return

    def isRequired(self) -> bool:
        """
        Return in this field a required one
        """
        return self.getTag("required")

    def setDesc(self, desc: str | None) -> None:
        """
        Set the description of this field
        """
        self.setTagData("desc", desc)

    def getDesc(self) -> str | None:
        """
        Return the description of this field
        """
        return self.getTagData("desc")

    def setValue(self, val) -> None:
        """
        Set the value of this field
        """
        self.setTagData("value", val)

    def getValue(self):
        return self.getTagData("value")

    def setValues(self, lst) -> None:
        """
        Set the values of this field as values-list. Replaces all previous filed
        values! If you need to just add a value - use addValue method
        """
        while self.getTag("value"):
            self.delChild("value")
        for val in lst:
            self.addValue(val)

    def addValue(self, val) -> None:
        """
        Add one more value to this field. Used in 'get' iq's or such
        """
        self.addChild("value", {}, [val])

    def getValues(self) -> list[str]:
        """
        Return the list of values associated with this field
        """
        ret: list[str] = []
        for tag in self.getTags("value"):
            ret.append(tag.getData())
        return ret

    def getOptions(self):
        """
        Return label-option pairs list associated with this field
        """
        ret = []
        for tag in self.getTags("option"):
            ret.append([tag.getAttr("label"), tag.getTagData("value")])
        return ret

    def setOptions(self, lst: list[str]) -> None:
        """
        Set label-option pairs list associated with this field
        """
        while self.getTag("option"):
            self.delChild("option")
        for opt in lst:
            self.addOption(opt)

    def addOption(self, opt: list[str] | str) -> None:
        """
        Add one more label-option pair to this field
        """
        if isinstance(opt, list):
            self.addChild("option", {"label": opt[0]}).setTagData("value", opt[1])
        else:
            self.addChild("option").setTagData("value", opt)

    def getType(self):
        """
        Get type of this field
        """
        return self.getAttr("type")

    def setType(self, val: str) -> None:
        """
        Set type of this field
        """
        return self.setAttr("type", val)

    def getVar(self):
        """
        Get 'var' attribute value of this field
        """
        return self.getAttr("var")

    def setVar(self, val: str) -> None:
        """
        Set 'var' attribute value of this field
        """
        return self.setAttr("var", val)


class DataForm(Node):
    """
    Used for manipulating dataforms in XMPP

    Relevant XEPs: 0004, 0068, 0122. Can be used in disco, pub-sub and many
    other applications.
    """

    def __init__(
        self,
        typ: str | None = None,
        data=None,
        title: str | None = None,
        node: Node | None = None,
    ) -> None:
        """
        Create new dataform of type 'typ'. 'data' is the list of DataField
        instances that this dataform contains, 'title' - the title string.  You
        can specify the 'node' argument as the other node to be used as base for
        constructing this dataform

        title and instructions is optional and SHOULD NOT contain newlines.
        Several instructions MAY be present.
        'typ' can be one of ('form' | 'submit' | 'cancel' | 'result' )
        'typ' of reply iq can be ( 'result' | 'set' | 'set' | 'result' )
            respectively.
        'cancel' form can not contain any fields. All other forms contains
            AT LEAST one field.
        'title' MAY be included in forms of type "form" and "result"
        """
        Node.__init__(self, "x", node=node)
        if node:
            newkids = []
            for n in self.getChildren():
                if n.getName() == "field":
                    newkids.append(DataField(node=n))
                else:
                    newkids.append(n)
            self.kids = newkids
        if typ:
            self.setType(typ)
        self.setNamespace(Namespace.DATA)
        if title:
            self.setTitle(title)
        if data is not None:
            if isinstance(data, dict):
                newdata = []
                for name in data:
                    newdata.append(DataField(name, data[name]))
                data = newdata
            for child in data:
                if child.__class__.__name__ == "DataField":
                    self.kids.append(child)
                elif isinstance(child, Node):
                    self.kids.append(DataField(node=child))
                else:  # Must be a string
                    self.addInstructions(child)

    def getType(self):
        """
        Return the type of dataform
        """
        return self.getAttr("type")

    def setType(self, typ) -> None:
        """
        Set the type of dataform
        """
        self.setAttr("type", typ)

    def getTitle(self):
        """
        Return the title of dataform
        """
        return self.getTagData("title")

    def setTitle(self, text: str) -> None:
        """
        Set the title of dataform
        """
        self.setTagData("title", text)

    def getInstructions(self):
        """
        Return the instructions of dataform
        """
        return self.getTagData("instructions")

    def setInstructions(self, text: str) -> None:
        """
        Set the instructions of dataform
        """
        self.setTagData("instructions", text)

    def addInstructions(self, text: str) -> None:
        """
        Add one more instruction to the dataform
        """
        self.addChild("instructions", {}, [text])

    def getField(self, name: str) -> Node | None:
        """
        Return the datafield object with name 'name' (if exists)
        """
        return self.getTag("field", attrs={"var": name})

    def setField(self, name: str):
        """
        Create if nessessary or get the existing datafield object with name
        'name' and return it
        """
        f = self.getField(name)
        if f:
            return f
        return self.addChild(node=DataField(name))

    def asDict(self):
        """
        Represent dataform as simple dictionary mapping of datafield names to
        their values
        """
        ret = {}
        for field in self.getTags("field"):
            name = field.getAttr("var")
            typ = field.getType()
            if typ and typ.endswith("-multi"):
                val = []
                for i in field.getTags("value"):
                    val.append(i.getData())
            else:
                val = field.getTagData("value")
            ret[name] = val
        if self.getTag("instructions"):
            ret["instructions"] = self.getInstructions()
        return ret

    def __getitem__(self, name: str):
        """
        Simple dictionary interface for getting datafields values by their names
        """
        item = self.getField(name)
        if item:
            return item.getValue()
        raise IndexError("No such field")

    def __setitem__(self, name: str, val):
        """
        Simple dictionary interface for setting datafields values by their names
        """
        return self.setField(name).setValue(val)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

import datetime as dt
import logging

from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import Presence
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client

log = logging.getLogger("nbxmpp.m.delay")


class Delay(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_delay,
                ns=Namespace.DELAY2,
                priority=15,
            ),
            StanzaHandler(
                name="presence",
                callback=self._process_presence_delay,
                ns=Namespace.DELAY2,
                priority=15,
            ),
        ]

    def _process_message_delay(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        # Determine if delay is from the server
        # Some servers use the bare jid, others the domain
        our_jid = self._client.get_bound_jid()
        jids = [our_jid.bare, our_jid.domain]

        if properties.from_muc:
            muc_jid = properties.jid
            jids += [muc_jid.bare, muc_jid.domain]

        if properties.is_muc_subject:
            # MUC Subjects can have a delay timestamp
            # to indicate when the user has set the subject,
            # the 'from' attr on these delays is the MUC server
            # but we treat it as user timestamp
            properties.user_timestamp = parse_delay(stanza, from_=jids)

        else:
            server_delay = parse_delay(stanza, from_=jids)
            if server_delay is not None:
                properties.has_server_delay = True
                properties.timestamp = server_delay

            properties.user_timestamp = parse_delay(stanza, not_from=jids)

    @staticmethod
    def _process_presence_delay(
        _client: Client, stanza: Presence, properties: PresenceProperties
    ) -> None:
        properties.user_timestamp = parse_delay(stanza)


def parse_delay(
    stanza: Message | Presence,
    epoch: bool = True,
    convert: str = "utc",
    from_: list[str] | None = None,
    not_from: list[str] | None = None,
) -> dt.datetime | float | None:
    """
    Returns the first valid delay timestamp that matches

    :param epoch:      Returns the timestamp as epoch

    :param convert:    Converts the timestamp to either utc or local

    :param from_:      Matches only delays that have the according
                       from attr set

    :param not_from:   Matches only delays that have the according
                       from attr not set
    """
    delays = stanza.getTags("delay", namespace=Namespace.DELAY2)

    for delay in delays:
        stamp = delay.getAttr("stamp")
        if stamp is None:
            log.warning("Invalid timestamp received: %s", stamp)
            log.warning(stanza)
            continue

        delay_from = delay.getAttr("from")
        if from_ is not None:
            if delay_from not in from_:
                continue
        if not_from is not None:
            if delay_from in not_from:
                continue

        timestamp = parse_datetime(stamp, check_utc=True, epoch=epoch, convert=convert)
        if timestamp is None:
            log.warning("Invalid timestamp received: %s", stamp)
            log.warning(stanza)
            continue

        return timestamp



from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp import Namespace
from nbxmpp import Node
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import RetractionData
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Retraction(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message,
                ns=Namespace.MESSAGE_RETRACT_1,
                priority=20,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_message_retracted_tombstone,
                ns=Namespace.MESSAGE_RETRACT_1,
                priority=20,
            ),
        ]

    def _process_message(
        self, _client: Client, stanza: Node, properties: MessageProperties
    ) -> None:
        retraction = stanza.getTag("retract", namespace=Namespace.MESSAGE_RETRACT_1)

        if retraction is None:
            return

        retracted_id = retraction.getAttr("id")

        if retracted_id is None:
            self._log.warning("<retract> without id")
            return

        properties.retraction = RetractionData(
            id=retracted_id, is_tombstone=False, timestamp=None
        )

    def _process_message_retracted_tombstone(
        self, _client: Client, stanza: Node, properties: MessageProperties
    ) -> None:
        if not properties.is_mam_message:
            return

        retracted = stanza.getTag("retracted", namespace=Namespace.MESSAGE_RETRACT_1)

        if retracted is None:
            return

        retracted_id = retracted.getAttr("id")

        if retracted_id is None:
            self._log.warning("<retracted> without id")
            return

        retracted_stamp = retracted.getAttr("stamp")

        properties.retraction = RetractionData(
            id=retracted_id,
            is_tombstone=True,
            timestamp=parse_datetime(
                retracted_stamp, check_utc=True, convert="utc", epoch=True
            ),
        )



import unittest

from nbxmpp.modules.discovery import parse_disco_info
from nbxmpp.protocol import DiscoInfoMalformed
from nbxmpp.protocol import Iq
from nbxmpp.util import compute_caps_hash


class EntityCaps(unittest.TestCase):

    def test_multiple_field_values(self):
        node = """
        <iq from='benvolio@capulet.lit/230193' id='disco1' to='juliet@capulet.lit/chamber' type='result'>
          <query xmlns='http://jabber.org/protocol/disco#info' node='http://psi-im.org#q07IKJEyjvHSyhy//CH0CxmKi8w='>
            <identity xml:lang='en' category='client' name='Psi 0.11' type='pc'/>
            <identity xml:lang='el' category='client' name='Ψ 0.11' type='pc'/>
            <feature var='http://jabber.org/protocol/caps'/>
            <feature var='http://jabber.org/protocol/disco#info'/>
            <feature var='http://jabber.org/protocol/disco#items'/>
            <feature var='http://jabber.org/protocol/muc'/>
            <x xmlns='jabber:x:data' type='result'>
              <field var='FORM_TYPE' type='hidden'>
                <value>urn:xmpp:dataforms:softwareinfo</value>
              </field>
              <field var='ip_version'>
                <value>ipv4</value>
                <value>ipv6</value>
              </field>
              <field var='os'>
                <value>Mac</value>
              </field>
              <field var='os_version'>
                <value>10.5.1</value>
              </field>
              <field var='software'>
                <value>Psi</value>
              </field>
              <field var='software_version'>
                <value>0.11</value>
              </field>
            </x>
          </query>
        </iq>"""

        info = parse_disco_info(Iq(node=node))
        hash_ = compute_caps_hash(info)
        self.assertEqual(hash_, "q07IKJEyjvHSyhy//CH0CxmKi8w=")

    def test_ignore_invalid_forms(self):
        node = """
        <iq from='benvolio@capulet.lit/230193' id='disco1' to='juliet@capulet.lit/chamber' type='result'>
          <query xmlns='http://jabber.org/protocol/disco#info' node='http://psi-im.org#q07IKJEyjvHSyhy//CH0CxmKi8w='>
            <identity xml:lang='en' category='client' name='Psi 0.11' type='pc'/>
            <identity xml:lang='el' category='client' name='Ψ 0.11' type='pc'/>
            <feature var='http://jabber.org/protocol/caps'/>
            <feature var='http://jabber.org/protocol/disco#info'/>
            <feature var='http://jabber.org/protocol/disco#items'/>
            <feature var='http://jabber.org/protocol/muc'/>
            <x xmlns='jabber:x:data' type='result'>
              <field var='FORM_TYPE' type='hidden'>
                <value>urn:xmpp:dataforms:softwareinfo</value>
              </field>
              <field var='ip_version'>
                <value>ipv4</value>
                <value>ipv6</value>
              </field>
              <field var='os'>
                <value>Mac</value>
              </field>
              <field var='os_version'>
                <value>10.5.1</value>
              </field>
              <field var='software'>
                <value>Psi</value>
              </field>
              <field var='software_version'>
                <value>0.11</value>
              </field>
            </x>
            <x xmlns='jabber:x:data' type='result'>
              <field var='FORM_TYPE'>
                <value>urn:xmpp:dataforms:softwareinfo</value>
              </field>
              <field var='ip_version'>
                <value>ipv4</value>
                <value>ipv6</value>
              </field>
            </x>
            <x xmlns='jabber:x:data' type='result'>
              <field var='ip_version'>
                <value>ipv4</value>
                <value>ipv6</value>
              </field>
            </x>
          </query>
        </iq>"""

        info = parse_disco_info(Iq(node=node))
        hash_ = compute_caps_hash(info)
        self.assertEqual(hash_, "q07IKJEyjvHSyhy//CH0CxmKi8w=")

    def test_multiple_form_type_values(self):
        node = """
        <iq from='benvolio@capulet.lit/230193' id='disco1' to='juliet@capulet.lit/chamber' type='result'>
          <query xmlns='http://jabber.org/protocol/disco#info' node='http://psi-im.org#q07IKJEyjvHSyhy//CH0CxmKi8w='>
            <identity xml:lang='en' category='client' name='Psi 0.11' type='pc'/>
            <identity xml:lang='el' category='client' name='Ψ 0.11' type='pc'/>
            <feature var='http://jabber.org/protocol/caps'/>
            <feature var='http://jabber.org/protocol/disco#info'/>
            <feature var='http://jabber.org/protocol/disco#items'/>
            <feature var='http://jabber.org/protocol/muc'/>
            <x xmlns='jabber:x:data' type='result'>
              <field var='FORM_TYPE' type='hidden'>
                <value>urn:xmpp:dataforms:softwareinfo</value>
                <value>urn:xmpp:dataforms:softwareinfo_test</value>
              </field>
              <field var='ip_version'>
                <value>ipv4</value>
                <value>ipv6</value>
              </field>
            </x>
          </query>
        </iq>"""

        info = parse_disco_info(Iq(node=node))
        with self.assertRaises(DiscoInfoMalformed):
            compute_caps_hash(info)

    def test_non_unique_form_type_value(self):
        node = """
        <iq from='benvolio@capulet.lit/230193' id='disco1' to='juliet@capulet.lit/chamber' type='result'>
          <query xmlns='http://jabber.org/protocol/disco#info' node='http://psi-im.org#q07IKJEyjvHSyhy//CH0CxmKi8w='>
            <identity xml:lang='en' category='client' name='Psi 0.11' type='pc'/>
            <identity xml:lang='el' category='client' name='Ψ 0.11' type='pc'/>
            <feature var='http://jabber.org/protocol/caps'/>
            <feature var='http://jabber.org/protocol/disco#info'/>
            <feature var='http://jabber.org/protocol/disco#items'/>
            <feature var='http://jabber.org/protocol/muc'/>
            <x xmlns='jabber:x:data' type='result'>
              <field var='FORM_TYPE' type='hidden'>
                <value>urn:xmpp:dataforms:softwareinfo</value>
              </field>
              <field var='ip_version'>
                <value>ipv4</value>
                <value>ipv6</value>
              </field>
            </x>
            <x xmlns='jabber:x:data' type='result'>
              <field var='FORM_TYPE' type='hidden'>
                <value>urn:xmpp:dataforms:softwareinfo</value>
              </field>
              <field var='ip_version'>
                <value>ipv4</value>
                <value>ipv6</value>
              </field>
            </x>
          </query>
        </iq>"""

        info = parse_disco_info(Iq(node=node))
        with self.assertRaises(DiscoInfoMalformed):
            compute_caps_hash(info)

    def test_non_unique_feature(self):
        node = """
        <iq from='benvolio@capulet.lit/230193' id='disco1' to='juliet@capulet.lit/chamber' type='result'>
          <query xmlns='http://jabber.org/protocol/disco#info' node='http://psi-im.org#q07IKJEyjvHSyhy//CH0CxmKi8w='>
            <identity xml:lang='en' category='client' name='Psi 0.11' type='pc'/>
            <identity xml:lang='el' category='client' name='Ψ 0.11' type='pc'/>
            <feature var='http://jabber.org/protocol/caps'/>
            <feature var='http://jabber.org/protocol/muc'/>
            <feature var='http://jabber.org/protocol/disco#info'/>
            <feature var='http://jabber.org/protocol/disco#items'/>
            <feature var='http://jabber.org/protocol/muc'/>
          </query>
        </iq>"""

        info = parse_disco_info(Iq(node=node))
        with self.assertRaises(DiscoInfoMalformed):
            compute_caps_hash(info)

    def test_non_unique_identity(self):
        node = """
        <iq from='benvolio@capulet.lit/230193' id='disco1' to='juliet@capulet.lit/chamber' type='result'>
          <query xmlns='http://jabber.org/protocol/disco#info' node='http://psi-im.org#q07IKJEyjvHSyhy//CH0CxmKi8w='>
            <identity xml:lang='en' category='client' name='Psi 0.11' type='pc'/>
            <identity xml:lang='en' category='client' name='Psi 0.11' type='pc'/>
            <feature var='http://jabber.org/protocol/caps'/>
            <feature var='http://jabber.org/protocol/muc'/>
            <feature var='http://jabber.org/protocol/disco#info'/>
            <feature var='http://jabber.org/protocol/disco#items'/>
            <feature var='http://jabber.org/protocol/muc'/>
          </query>
        </iq>"""

        info = parse_disco_info(Iq(node=node))
        with self.assertRaises(DiscoInfoMalformed):
            compute_caps_hash(info)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

import hashlib
from dataclasses import dataclass
from dataclasses import field

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.simplexml import Node
from nbxmpp.task import iq_request_task
from nbxmpp.util import b64decode
from nbxmpp.util import b64encode

if TYPE_CHECKING:
    from nbxmpp.client import Client


class VCardTemp(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_vcard(self, jid: JID | None = None):
        _task = yield

        response = yield _make_vcard_request(jid)

        if response.isError():
            raise StanzaError(response)

        vcard_node = _get_vcard_node(response)
        yield VCard.from_node(vcard_node)

    @iq_request_task
    def set_vcard(self, vcard: VCard, jid: JID | None = None):
        _task = yield

        response = yield _make_vcard_publish(jid, vcard)
        yield process_response(response)


def _make_vcard_request(jid: JID) -> Iq:
    iq = Iq(typ="get", to=jid)
    iq.addChild("vCard", namespace=Namespace.VCARD)
    return iq


def _get_vcard_node(response: Node) -> Node:
    vcard_node = response.getTag("vCard", namespace=Namespace.VCARD)
    if vcard_node is None:
        raise MalformedStanzaError("vCard node missing", response)
    return vcard_node


def _make_vcard_publish(jid: JID, vcard: VCard) -> Iq:
    iq = Iq(typ="set", to=jid)
    iq.addChild(node=vcard.to_node())
    return iq


@dataclass
class VCard:
    data: dict = field(default_factory=dict)

    @classmethod
    def from_node(cls, node: Node) -> VCard:
        dict_ = {}
        for info in node.getChildren():
            name = info.getName()
            if name in ("ADR", "TEL", "EMAIL"):
                dict_.setdefault(name, [])
                entry = {}
                for child in info.getChildren():
                    entry[child.getName()] = child.getData()
                dict_[name].append(entry)
            elif info.getChildren() == []:
                dict_[name] = info.getData()
            else:
                dict_[name] = {}
                for child in info.getChildren():
                    dict_[name][child.getName()] = child.getData()

        return cls(data=dict_)

    def to_node(self) -> Node:
        vcard = Node(tag="vCard", attrs={"xmlns": Namespace.VCARD})
        for i in self.data:
            if i == "jid":
                continue
            if isinstance(self.data[i], dict):
                child = vcard.addChild(i)
                for j in self.data[i]:
                    child.addChild(j).setData(self.data[i][j])
            elif isinstance(self.data[i], list):
                for j in self.data[i]:
                    child = vcard.addChild(i)
                    for k in j:
                        child.addChild(k).setData(j[k])
            else:
                vcard.addChild(i).setData(self.data[i])
        return vcard

    def set_avatar(self, avatar: bytes, type_: str | None = None) -> None:
        encoded_avatar = b64encode(avatar)
        if "PHOTO" not in self.data:
            self.data["PHOTO"] = {}

        self.data["PHOTO"]["BINVAL"] = encoded_avatar

        if type_ is not None:
            self.data["PHOTO"]["TYPE"] = type_

    def get_avatar(self) -> tuple[bytes | None, str | None]:
        try:
            avatar = self.data["PHOTO"]["BINVAL"]
        except Exception:
            return None, None

        if not avatar:
            return None, None

        encoded_avatar = b64decode(avatar)
        avatar_sha = hashlib.sha1(encoded_avatar).hexdigest()
        return encoded_avatar, avatar_sha



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass


@dataclass(frozen=True)
class _Namespaces:

    ACTIVITY: str = "http://jabber.org/protocol/activity"
    ADDRESS: str = "http://jabber.org/protocol/address"
    AGENTS: str = "jabber:iq:agents"
    ATTENTION: str = "urn:xmpp:attention:0"
    AUTH: str = "jabber:iq:auth"
    AVATAR_METADATA: str = "urn:xmpp:avatar:metadata"
    AVATAR_DATA: str = "urn:xmpp:avatar:data"
    BIND: str = "urn:ietf:params:xml:ns:xmpp-bind"
    BLOCKING: str = "urn:xmpp:blocking"
    BOB: str = "urn:xmpp:bob"
    BOOKMARKS: str = "storage:bookmarks"
    BOOKMARKS_1: str = "urn:xmpp:bookmarks:1"
    BOOKMARKS_COMPAT: str = "urn:xmpp:bookmarks:1#compat"
    BOOKMARKS_COMPAT_PEP: str = "urn:xmpp:bookmarks:1#compat-pep"
    BOOKMARK_CONVERSION: str = "urn:xmpp:bookmarks-conversion:0"
    BROWSE: str = "jabber:iq:browse"
    BYTESTREAM: str = "http://jabber.org/protocol/bytestreams"
    CAPS: str = "http://jabber.org/protocol/caps"
    CAPTCHA: str = "urn:xmpp:captcha"
    CARBONS: str = "urn:xmpp:carbons:2"
    CHANNEL_BINDING: str = "urn:xmpp:sasl-cb:0"
    CHATMARKERS: str = "urn:xmpp:chat-markers:0"
    CHATSTATES: str = "http://jabber.org/protocol/chatstates"
    CLIENT: str = "jabber:client"
    COMMANDS: str = "http://jabber.org/protocol/commands"
    CONFERENCE: str = "jabber:x:conference"
    CORRECT: str = "urn:xmpp:message-correct:0"
    DATA: str = "jabber:x:data"
    DATA_LAYOUT: str = "http://jabber.org/protocol/xdata-layout"
    DATA_MEDIA: str = "urn:xmpp:media-element"
    DATA_VALIDATE: str = "http://jabber.org/protocol/xdata-validate"
    DELAY: str = "jabber:x:delay"
    DELAY2: str = "urn:xmpp:delay"
    DELIMITER: str = "roster:delimiter"
    DISCO: str = "http://jabber.org/protocol/disco"
    DISCO_INFO: str = "http://jabber.org/protocol/disco#info"
    DISCO_ITEMS: str = "http://jabber.org/protocol/disco#items"
    DOMAIN_BASED_NAME: str = "urn:xmpp:domain-based-name:1"
    EME: str = "urn:xmpp:eme:0"
    ENCRYPTED: str = "jabber:x:encrypted"
    FALLBACK: str = "urn:xmpp:fallback:0"
    FASTEN: str = "urn:xmpp:fasten:0"
    FILE_METADATA: str = "urn:xmpp:file:metadata:0"
    FORWARD: str = "urn:xmpp:forward:0"
    FRAMING: str = "urn:ietf:params:xml:ns:xmpp-framing"
    GATEWAY: str = "jabber:iq:gateway"
    GEOLOC: str = "http://jabber.org/protocol/geoloc"
    HASHES: str = "urn:xmpp:hashes:1"
    HASHES_2: str = "urn:xmpp:hashes:2"
    HASHES_BLAKE2B_256: str = "urn:xmpp:hash-function-text-names:id-blake2b256"
    HASHES_BLAKE2B_512: str = "urn:xmpp:hash-function-text-names:id-blake2b512"
    HASHES_MD5: str = "urn:xmpp:hash-function-text-names:md5"
    HASHES_SHA1: str = "urn:xmpp:hash-function-text-names:sha-1"
    HASHES_SHA256: str = "urn:xmpp:hash-function-text-names:sha-256"
    HASHES_SHA512: str = "urn:xmpp:hash-function-text-names:sha-512"
    HASHES_SHA3_256: str = "urn:xmpp:hash-function-text-names:sha3-256"
    HASHES_SHA3_512: str = "urn:xmpp:hash-function-text-names:sha3-512"
    HATS: str = "urn:xmpp:hats:0"
    HATS_LEGACY: str = "xmpp:prosody.im/protocol/hats:1"
    HINTS: str = "urn:xmpp:hints"
    HTTPUPLOAD_0: str = "urn:xmpp:http:upload:0"
    HTTP_AUTH: str = "http://jabber.org/protocol/http-auth"
    IBB: str = "http://jabber.org/protocol/ibb"
    IDLE: str = "urn:xmpp:idle:1"
    JINGLE: str = "urn:xmpp:jingle:1"
    JINGLE_BYTESTREAM: str = "urn:xmpp:jingle:transports:s5b:1"
    JINGLE_DTLS: str = "urn:xmpp:jingle:apps:dtls:0"
    JINGLE_ERRORS: str = "urn:xmpp:jingle:errors:1"
    JINGLE_FILE_TRANSFER: str = "urn:xmpp:jingle:apps:file-transfer:3"
    JINGLE_FILE_TRANSFER_5: str = "urn:xmpp:jingle:apps:file-transfer:5"
    JINGLE_IBB: str = "urn:xmpp:jingle:transports:ibb:1"
    JINGLE_ICE_UDP: str = "urn:xmpp:jingle:transports:ice-udp:1"
    JINGLE_RAW_UDP: str = "urn:xmpp:jingle:transports:raw-udp:1"
    JINGLE_RTCP_FB: str = "urn:xmpp:jingle:apps:rtp:rtcp-fb:0"
    JINGLE_RTP: str = "urn:xmpp:jingle:apps:rtp:1"
    JINGLE_RTP_AUDIO: str = "urn:xmpp:jingle:apps:rtp:audio"
    JINGLE_RTP_VIDEO: str = "urn:xmpp:jingle:apps:rtp:video"
    JINGLE_XTLS: str = "urn:xmpp:jingle:security:xtls:0"
    JINGLE_MESSAGE: str = "urn:xmpp:jingle-message:0"
    LAST: str = "jabber:iq:last"
    LOCATION: str = "http://jabber.org/protocol/geoloc"
    MAM_1: str = "urn:xmpp:mam:1"
    MAM_2: str = "urn:xmpp:mam:2"
    MESSAGE_MODERATE: str = "urn:xmpp:message-moderate:0"
    MESSAGE_MODERATE_1: str = "urn:xmpp:message-moderate:1"
    MESSAGE_RETRACT: str = "urn:xmpp:message-retract:0"
    MESSAGE_RETRACT_1: str = "urn:xmpp:message-retract:1"
    MDS: str = "urn:xmpp:mds:displayed:0"
    MOOD: str = "http://jabber.org/protocol/mood"
    MSG_HINTS: str = "urn:xmpp:hints"
    MUCLUMBUS: str = "https://xmlns.zombofant.net/muclumbus/search/1.0"
    MUC: str = "http://jabber.org/protocol/muc"
    MUC_USER: str = "http://jabber.org/protocol/muc#user"
    MUC_ADMIN: str = "http://jabber.org/protocol/muc#admin"
    MUC_OWNER: str = "http://jabber.org/protocol/muc#owner"
    MUC_UNIQUE: str = "http://jabber.org/protocol/muc#unique"
    MUC_CONFIG: str = "http://jabber.org/protocol/muc#roomconfig"
    MUC_REQUEST: str = "http://jabber.org/protocol/muc#request"
    MUC_INFO: str = "http://jabber.org/protocol/muc#roominfo"
    NICK: str = "http://jabber.org/protocol/nick"
    OCCUPANT_ID: str = "urn:xmpp:occupant-id:0"
    OMEMO_TEMP: str = "eu.siacs.conversations.axolotl"
    OMEMO_TEMP_BUNDLE: str = "eu.siacs.conversations.axolotl.bundles"
    OMEMO_TEMP_DL: str = "eu.siacs.conversations.axolotl.devicelist"
    OPENPGP: str = "urn:xmpp:openpgp:0"
    OPENPGP_PK: str = "urn:xmpp:openpgp:0:public-keys"
    OPENPGP_SK: str = "urn:xmpp:openpgp:0:secret-key"
    PING: str = "urn:xmpp:ping"
    PRIVACY: str = "jabber:iq:privacy"
    PRIVATE: str = "jabber:iq:private"
    PUBKEY_ATTEST: str = "urn:xmpp:attest:2"
    PUBKEY_PUBKEY: str = "urn:xmpp:pubkey:2"
    PUBKEY_REVOKE: str = "urn:xmpp:revoke:2"
    PUBSUB: str = "http://jabber.org/protocol/pubsub"
    PUBSUB_ERROR: str = "http://jabber.org/protocol/pubsub#errors"
    PUBSUB_CONFIG: str = "http://jabber.org/protocol/pubsub#node_config"
    PUBSUB_EVENT: str = "http://jabber.org/protocol/pubsub#event"
    PUBSUB_OWNER: str = "http://jabber.org/protocol/pubsub#owner"
    PUBSUB_PUBLISH_OPTIONS: str = "http://jabber.org/protocol/pubsub#publish-options"
    PUBSUB_NODE_MAX: str = "http://jabber.org/protocol/pubsub#config-node-max"
    REACTIONS: str = "urn:xmpp:reactions:0"
    RECEIPTS: str = "urn:xmpp:receipts"
    REGISTER: str = "jabber:iq:register"
    REGISTER_FEATURE: str = "http://jabber.org/features/iq-register"
    REPLY: str = "urn:xmpp:reply:0"
    REPORTING: str = "urn:xmpp:reporting:0"
    ROSTER: str = "jabber:iq:roster"
    ROSTERNOTES: str = "storage:rosternotes"
    ROSTERX: str = "http://jabber.org/protocol/rosterx"
    ROSTER_VER: str = "urn:xmpp:features:rosterver"
    RSM: str = "http://jabber.org/protocol/rsm"
    SASL: str = "urn:ietf:params:xml:ns:xmpp-sasl"
    SASL2: str = "urn:xmpp:sasl:2"
    SEARCH: str = "jabber:iq:search"
    SECLABEL: str = "urn:xmpp:sec-label:0"
    SECLABEL_CATALOG: str = "urn:xmpp:sec-label:catalog:2"
    SERVICE_DISCOVERY: str = "urn:xmpp:extdisco:2"
    SESSION: str = "urn:ietf:params:xml:ns:xmpp-session"
    SFS: str = "urn:xmpp:sfs:0"
    SID: str = "urn:xmpp:sid:0"
    SIGNED: str = "jabber:x:signed"
    SIMS: str = "urn:xmpp:sims:1"
    STANZAS: str = "urn:ietf:params:xml:ns:xmpp-stanzas"
    STICKERS: str = "urn:xmpp:stickers:0"
    STREAM: str = "http://affinix.com/jabber/stream"
    STREAMS: str = "http://etherx.jabber.org/streams"
    STREAM_MGMT: str = "urn:xmpp:sm:3"
    STYLING: str = "urn:xmpp:styling:0"
    TIME_REVISED: str = "urn:xmpp:time"
    TIME: str = "urn:xmpp:time"
    TLS: str = "urn:ietf:params:xml:ns:xmpp-tls"
    TUNE: str = "http://jabber.org/protocol/tune"
    URL_DATA: str = "http://jabber.org/protocol/url-data"
    VCARD: str = "vcard-temp"
    VCARD_UPDATE: str = "vcard-temp:x:update"
    VCARD_CONVERSION: str = "urn:xmpp:pep-vcard-conversion:0"
    VCARD4: str = "urn:ietf:params:xml:ns:vcard-4.0"
    VCARD4_PUBSUB: str = "urn:xmpp:vcard4"
    VERSION: str = "jabber:iq:version"
    XHTML_IM: str = "http://jabber.org/protocol/xhtml-im"
    XHTML: str = "http://www.w3.org/1999/xhtml"
    XMPP_STREAMS: str = "urn:ietf:params:xml:ns:xmpp-streams"
    X_OOB: str = "jabber:x:oob"
    XRD: str = "http://docs.oasis-open.org/ns/xri/xrd-1.0"


Namespace: _Namespaces = _Namespaces()



from __future__ import annotations

from test.lib.util import StanzaHandlerTest
from unittest.mock import MagicMock

import nbxmpp
from nbxmpp.exceptions import FallbackLanguageError
from nbxmpp.language import LanguageRange
from nbxmpp.modules.fallback import FallbackRange
from nbxmpp.modules.fallback import FallbacksForT
from nbxmpp.modules.fallback import parse_fallback_indication
from nbxmpp.modules.fallback import strip_fallback
from nbxmpp.structs import BodyData


class TestFallback(StanzaHandlerTest):

    def test_parse_fallback_indication(self):
        xml = """
            <message to='anna@example.com' id='message-id2' type='groupchat'>
              <body>> Anna wrote:\n> Hi, how are you?\nGreat</body>
              <fallback xmlns='urn:xmpp:fallback:0' for='urn:xmpp:test:0'>
                <body start='0' end='33' />
                <body xml:lang='en' start='0' end='32' />
                <body xml:lang='de' start='0' end='31' />
              </fallback>
              <fallback xmlns='urn:xmpp:fallback:0' for='urn:xmpp:test:1'>
                <body />
              </fallback>
              <fallback xmlns='urn:xmpp:fallback:0' for='urn:xmpp:test:2' />
            </message>
        """

        log = MagicMock()
        message = nbxmpp.Message(node=xml)

        fallbacks_for = parse_fallback_indication(log, message)

        expected = {
            "urn:xmpp:test:0": {
                None: FallbackRange(0, 33),
                "en": FallbackRange(0, 32),
                "de": FallbackRange(0, 31),
            },
            "urn:xmpp:test:1": {
                None: None,
            },
            "urn:xmpp:test:2": None,
        }

        self.assertEqual(fallbacks_for, expected)

    def test_strip_fallback(self):
        fallbacks_for: FallbacksForT = {
            "urn:xmpp:test:1": {
                None: FallbackRange(33, 38),
            },
            "urn:xmpp:test:0": {
                None: FallbackRange(0, 14),
                "en": FallbackRange(0, 33),
            },
            "urn:xmpp:test:2": {
                None: FallbackRange(54, 55),
                "en": None,
            },
        }

        text = "> Anna wrote:\n> Hi, how are you?\nGreat"

        # Strip one fallback with unavailable language
        with self.assertRaises(FallbackLanguageError):
            stripped_text = strip_fallback(
                fallbacks_for, {"urn:xmpp:test:0"}, "xx", text
            )

        # Strip one fallback with language
        stripped_text = strip_fallback(fallbacks_for, {"urn:xmpp:test:0"}, "en", text)
        self.assertEqual(stripped_text, "Great")

        # Strip one fallback with no language
        stripped_text = strip_fallback(fallbacks_for, {"urn:xmpp:test:0"}, None, text)
        self.assertEqual(stripped_text, "> Hi, how are you?\nGreat")

        # Strip multiple fallbacks in correct order
        stripped_text = strip_fallback(
            fallbacks_for, {"urn:xmpp:test:0", "urn:xmpp:test:1"}, None, text
        )
        self.assertEqual(stripped_text, "> Hi, how are you?\n")

        # Range out of bounds, don’t strip anything
        stripped_text = strip_fallback(fallbacks_for, {"urn:xmpp:test:2"}, None, text)
        self.assertEqual(stripped_text, text)

        # One Range is None, so whole body is fallback
        stripped_text = strip_fallback(
            fallbacks_for, {"urn:xmpp:test:0", "urn:xmpp:test:2"}, "en", text
        )
        self.assertEqual(stripped_text, "")

    def test_body_with_fallback(self):
        text = "> Anna wrote:\n> Hi, how are you?\nGreat"
        text_de = "> Anna schrieb\n>Hallo, wie geht es dir?\nGut"

        xml = """
            <message to='anna@example.com' id='message-id2' type='groupchat'>
              <body>%s</body>
              <body xml:lang='en'>%s</body>
              <body xml:lang='de'>%s</body>
              <fallback xmlns='urn:xmpp:fallback:0' for='urn:xmpp:test:0'>
                <body start='0' end='33' />
                <body xml:lang='en' start='0' end='33' />
                <body xml:lang='de' start='0' end='40' />
              </fallback>
              <fallback xmlns='urn:xmpp:fallback:0' for='urn:xmpp:test:1'>
                <body />
              </fallback>
              <fallback xmlns='urn:xmpp:fallback:0' for='urn:xmpp:test:2' />
            </message>
        """ % (
            text,
            text,
            text_de,
        )

        log = MagicMock()
        message = nbxmpp.Message(node=xml)

        fallbacks_for = parse_fallback_indication(log, message)

        body_data = BodyData(message, fallbacks_for, {"urn:xmpp:test:0"})
        body = body_data.get([LanguageRange(tag="en")])
        self.assertEqual(body, "Great")

        body = body_data.get(None)
        self.assertEqual(body, "Great")

        body = body_data.get([LanguageRange(tag="de")])
        self.assertEqual(body, "Gut")

        body_data = BodyData(message, fallbacks_for, {"urn:xmpp:test:1"})
        # specific language body requested
        # but there is no fallback for this language
        body = body_data.get([LanguageRange(tag="en")])
        self.assertEqual(body, text)

        body = body_data.get(None)
        self.assertEqual(body, "")

        body_data = BodyData(message, fallbacks_for, {"urn:xmpp:test:2"})
        # specific language body requested
        # fallback is for all bodies
        body = body_data.get([LanguageRange(tag="en")])
        self.assertEqual(body, "")

        body = body_data.get(None)
        self.assertEqual(body, "")



from test.lib.util import StanzaHandlerTest

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import ActivityData
from nbxmpp.structs import PubSubEventData
from nbxmpp.structs import StanzaHandler


class ActivityTest(StanzaHandlerTest):

    def test_activity_parsing(self):
        def _on_message(_con, _stanza, properties):

            data = ActivityData(
                activity="relaxing", subactivity="partying", text="My nurse's birthday!"
            )

            pubsub_event = PubSubEventData(
                node="http://jabber.org/protocol/activity",
                id="b5ac48d0-0f9c-11dc-8754-001143d5d5db",
                item=None,
                data=data,
                deleted=False,
                retracted=False,
                purged=False,
            )

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = """
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='http://jabber.org/protocol/activity'>
                        <item id='b5ac48d0-0f9c-11dc-8754-001143d5d5db'>
                            <activity xmlns='http://jabber.org/protocol/activity'>
                                <relaxing>
                                    <partying/>
                                </relaxing>
                                <text xml:lang='en'>My nurse&apos;s birthday!</text>
                            </activity>
                        </item>
                    </items>
                </event>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message", callback=_on_message, ns=Namespace.PUBSUB_EVENT
            )
        )

        self.dispatcher.process_data(event)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.bookmarks.util import build_conference_node
from nbxmpp.modules.bookmarks.util import parse_bookmark
from nbxmpp.modules.util import finalize
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import BookmarkData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client

BOOKMARK_OPTIONS = {
    "pubsub#notify_delete": "true",
    "pubsub#notify_retract": "true",
    "pubsub#persist_items": "true",
    "pubsub#max_items": "max",
    "pubsub#access_model": "whitelist",
    "pubsub#send_last_published_item": "never",
}


class NativeBookmarks(BaseModule):

    _depends = {
        "retract": "PubSub",
        "publish": "PubSub",
        "request_items": "PubSub",
    }

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pubsub_bookmarks,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            ),
        ]

    def _process_pubsub_bookmarks(
        self, _client: Client, _stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.BOOKMARKS_1:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        try:
            bookmark_item = parse_bookmark(item)
        except MalformedStanzaError as error:
            self._log.warning(error)
            self._log.warning(error.stanza)
            raise NodeProcessed

        pubsub_event = properties.pubsub_event._replace(data=bookmark_item)
        self._log.info("Received bookmark item from: %s", properties.jid)
        self._log.info(bookmark_item)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def request_bookmarks(self):
        _task = yield

        items = yield self.request_items(Namespace.BOOKMARKS_1)
        raise_if_error(items)

        bookmarks: list[BookmarkData] = []
        for item in items:
            try:
                bookmark_item = parse_bookmark(item)
            except MalformedStanzaError as error:
                self._log.warning(error)
                self._log.warning(error.stanza)
                continue

            bookmarks.append(bookmark_item)

        for bookmark in bookmarks:
            self._log.info(bookmark)

        yield bookmarks

    @iq_request_task
    def retract_bookmark(self, bookmark_jid: JID):
        task = yield

        self._log.info("Retract Bookmark: %s", bookmark_jid)

        result = yield self.retract(Namespace.BOOKMARKS_1, str(bookmark_jid))
        yield finalize(task, result)

    @iq_request_task
    def store_bookmarks(self, bookmarks: list[BookmarkData]):
        _task = yield

        self._log.info("Store Bookmarks")

        for bookmark in bookmarks:
            self.publish(
                Namespace.BOOKMARKS_1,
                build_conference_node(bookmark),
                id_=str(bookmark.jid),
                options=BOOKMARK_OPTIONS,
                force_node_options=True,
            )

        yield True



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import TYPE_CHECKING

import base64
import binascii
import hashlib
import logging
import os
import re
import uuid
import xml.etree.ElementTree as ET
from collections import defaultdict
from collections.abc import Callable
from functools import lru_cache
from logging import LoggerAdapter

from gi.repository import Gio
from gi.repository import Soup
from packaging.version import Version

from nbxmpp.const import GIO_TLS_ERRORS
from nbxmpp.const import GLIB_VERSION
from nbxmpp.const import SOUP_ENCODING
from nbxmpp.modules.dataforms import DataForm
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import DiscoInfoMalformed
from nbxmpp.protocol import JID
from nbxmpp.protocol import StanzaMalformed
from nbxmpp.protocol import StreamHeader
from nbxmpp.protocol import WebsocketOpenHeader
from nbxmpp.simplexml import Node
from nbxmpp.structs import CommonError
from nbxmpp.structs import DiscoIdentity
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import HTTPUploadError
from nbxmpp.structs import IqProperties
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import Properties
from nbxmpp.structs import StanzaMalformedError
from nbxmpp.third_party import hsluv

if TYPE_CHECKING:
    from nbxmpp.protocol import Protocol

log = logging.getLogger("nbxmpp.util")


def b64decode(data: str | bytes) -> bytes:
    if not data:
        raise ValueError("No data to decode")

    if isinstance(data, str):
        data = data.encode()

    return base64.b64decode(data)


def b64encode(data: str | bytes) -> str:
    if not data:
        raise ValueError("No data to encode")

    if isinstance(data, str):
        data = data.encode()

    result = base64.b64encode(data)
    return result.decode()


def get_properties_struct(
    name: str, own_jid: JID
) -> MessageProperties | IqProperties | PresenceProperties | Properties:
    if name == "message":
        return MessageProperties(own_jid)
    if name == "iq":
        return IqProperties(own_jid)
    if name == "presence":
        return PresenceProperties(own_jid)
    return Properties()


def from_xs_boolean(value: str) -> bool:
    if value in ("1", "true", "True"):
        return True

    if value in ("0", "false", "False", ""):
        return False

    raise ValueError("Cant convert %s to python boolean" % value)


def to_xs_boolean(value: bool | None) -> Literal["true", "false"]:
    # Convert to xs:boolean ('true', 'false')
    # from a python boolean (True, False) or None
    if value is True:
        return "true"

    if value is False:
        return "false"

    if value is None:
        return "false"

    raise ValueError("Cant convert %s to xs:boolean" % value)


error_classes = {Namespace.HTTPUPLOAD_0: HTTPUploadError}


def error_factory(
    stanza: Protocol, condition: str | None = None, text: str | None = None
) -> Any:
    if condition == "stanza-malformed":
        return StanzaMalformedError(stanza, text)
    app_namespace = stanza.getAppErrorNamespace()
    return error_classes.get(app_namespace, CommonError)(stanza)


def clip_rgb(red: float, green: float, blue: float) -> tuple[float, float, float]:
    return (
        min(max(red, 0), 1),
        min(max(green, 0), 1),
        min(max(blue, 0), 1),
    )


def text_to_hue(text: str) -> float:
    hash_ = hashlib.sha1()
    hash_.update(text.encode())
    return int.from_bytes(hash_.digest()[:2], "little") / 65536 * 360


def hsluv_to_rgb(
    hue: float,
    saturation: float,
    lightness: float,
) -> tuple[float, float, float]:
    return clip_rgb(*hsluv.hsluv_to_rgb((hue, saturation, lightness)))  # type: ignore


@lru_cache(maxsize=1024)
def text_to_color(
    text: str,
    saturation: float,
    lightness: float,
) -> tuple[float, float, float]:
    """
    Combines algorithms from XEP-0392 (version 0.8.0
    <https://xmpp.org/extensions/attic/xep-0392-0.8.0.html>):
        5.1 Angle generation,
        5.4 RGB generation.

    Input:
        text: the subject to compute hue for,
        saturation ∈ [0, 100],
        lightness ∈ [0, 100].

    Returns color in sRGB color space: r, g, b ∈ [0, 1].
    """
    return hsluv_to_rgb(text_to_hue(text), saturation, lightness)


def compute_caps_hash(info: DiscoInfo, compare: bool = True) -> str:
    """
    Compute caps hash according to XEP-0115, V1.5
    https://xmpp.org/extensions/xep-0115.html#ver-proc

    :param: info    DiscoInfo
    :param: compare If True an exception is raised if the hash announced in
                    the node attr is not equal to what is calculated
    """
    # Initialize an empty string S.
    string_ = ""

    # Sort the service discovery identities by category and then by type and
    # then by xml:lang (if it exists), formatted as
    # CATEGORY '/' [TYPE] '/' [LANG] '/' [NAME]. Note that each slash is
    # included even if the LANG or NAME is not included (in accordance with
    # XEP-0030, the category and type MUST be included).
    # For each identity, append the 'category/type/lang/name' to S, followed by
    # the '<' character.
    # Sort the supported service discovery features.

    def sort_identities_key(i: DiscoIdentity) -> tuple[str, str, str]:
        return (i.category, i.type, i.lang or "")

    identities = sorted(info.identities, key=sort_identities_key)
    for identity in identities:
        string_ += "%s<" % str(identity)

    # If the response includes more than one service discovery identity with
    # the same category/type/lang/name, consider the entire response
    # to be ill-formed.
    if len(set(identities)) != len(identities):
        raise DiscoInfoMalformed("Non-unique identity found")

    # Sort the supported service discovery features.
    # For each feature, append the feature to S, followed by the '<' character.
    features = sorted(info.features)
    for feature in features:
        string_ += "%s<" % feature

    # If the response includes more than one service discovery feature with the
    # same XML character data, consider the entire response to be ill-formed.
    if len(set(features)) != len(features):
        raise DiscoInfoMalformed("Non-unique feature found")

    # If the response includes more than one extended service discovery
    # information form with the same FORM_TYPE or the FORM_TYPE field contains
    # more than one <value/> element with different XML character data,
    # consider the entire response to be ill-formed.

    # If the response includes an extended service discovery information form
    # where the FORM_TYPE field is not of type "hidden" or the form does not
    # include a FORM_TYPE field, ignore the form but continue processing.

    dataforms: list[DataForm] = []
    form_type_values: list[Any] = []
    for dataform in info.dataforms:
        form_type = dataform.vars.get("FORM_TYPE")
        if form_type is None:
            # Ignore dataform because of missing FORM_TYPE
            continue
        if form_type.type_ != "hidden":
            # Ignore dataform because of wrong type
            continue

        values = form_type.getTags("value")
        if len(values) != 1:
            raise DiscoInfoMalformed("Form should have exactly " "one FORM_TYPE value")
        value = values[0].getData()

        dataforms.append(dataform)
        form_type_values.append(value)

    if len(set(form_type_values)) != len(form_type_values):
        raise DiscoInfoMalformed("Non-unique FORM_TYPE value found")

    # If the service discovery information response includes XEP-0128 data
    # forms, sort the forms by the FORM_TYPE (i.e., by the XML character data
    # of the <value/> element).

    # For each extended service discovery information form:
    #   - Append the XML character data of the FORM_TYPE field's <value/>
    #     element, followed by the '<' character.
    #   - Sort the fields by the value of the "var" attribute.
    #   - For each field other than FORM_TYPE:
    #       - Append the value of the "var" attribute, followed by the
    #         '<' character.
    #       - Sort values by the XML character data of the <value/> element.
    #       - For each <value/> element, append the XML character data,
    #         followed by the '<' character.

    def sort_dataforms_key(dataform: DataForm):
        return dataform["FORM_TYPE"].getTagData("value")

    dataforms = sorted(dataforms, key=sort_dataforms_key)
    for dataform in dataforms:
        string_ += "%s<" % dataform["FORM_TYPE"].getTagData("value")

        fields = {}
        for field in dataform.iter_fields():
            if field.var == "FORM_TYPE":
                continue
            values = field.getTags("value")
            fields[field.var] = sorted([value.getData() for value in values])

        for var in sorted(fields.keys()):
            string_ += "%s<" % var
            for value in fields[var]:
                string_ += "%s<" % value

    hash_ = hashlib.sha1(string_.encode())
    b64hash = b64encode(hash_.digest())
    if compare and b64hash != info.get_caps_hash():
        raise DiscoInfoMalformed(
            "Caps hashes differ: %s != %s" % (b64hash, info.get_caps_hash())
        )
    return b64hash


def generate_id() -> str:
    return str(uuid.uuid4())


def get_form(stanza: Node, form_type: Any) -> Any:
    forms = stanza.getTags("x", namespace=Namespace.DATA)
    if not forms:
        return None

    for form in forms:
        form = extend_form(node=form)
        field = form.vars.get("FORM_TYPE")
        if field is None:
            continue

        if field.value == form_type:
            return form
    return None


def validate_stream_header(stanza: Node, domain: str, is_websocket: bool) -> str:
    attrs = stanza.getAttrs()
    if attrs.get("from") != domain:
        raise StanzaMalformed("Invalid from attr in stream header")

    if is_websocket:
        if attrs.get("xmlns") != Namespace.FRAMING:
            raise StanzaMalformed("Invalid namespace in stream header")
    else:
        if attrs.get("xmlns:stream") != Namespace.STREAMS:
            raise StanzaMalformed("Invalid stream namespace in stream header")
        if attrs.get("xmlns") != Namespace.CLIENT:
            raise StanzaMalformed("Invalid namespace in stream header")

    if attrs.get("version") != "1.0":
        raise StanzaMalformed("Invalid stream version in stream header")
    stream_id = attrs.get("id")
    if stream_id is None:
        raise StanzaMalformed("No stream id found in stream header")
    return stream_id


def get_stream_header(domain: str, lang: str, is_websocket: bool) -> str:
    if is_websocket:
        return str(WebsocketOpenHeader(domain, lang))
    header = StreamHeader(domain, lang)
    return "<?xml version='1.0'?>%s>" % str(header)[:-3]


def utf8_decode(data: bytes) -> tuple[str, bytes]:
    """
    Decodes utf8 byte string to unicode string
    Does handle invalid utf8 sequences by splitting
    the invalid sequence at the end

    returns (decoded unicode string, invalid byte sequence)
    """
    try:
        return data.decode(), b""
    except UnicodeDecodeError:
        for i in range(-1, -4, -1):
            char = data[i]
            if char & 0xC0 == 0x80:
                continue
            return data[:i].decode(), data[i:]
        raise


def get_rand_number() -> int:
    return int(binascii.hexlify(os.urandom(6)), 16)


def get_invalid_xml_regex() -> re.Pattern[str]:
    # \ufddo -> \ufdef range
    c = "\ufdd0"
    r = c
    while c < "\ufdef":
        c = chr(ord(c) + 1)
        r += "|" + c

    # \ufffe-\uffff, \u1fffe-\u1ffff, ..., \u10fffe-\u10ffff
    c = "\ufffe"
    r += "|" + c
    r += "|" + chr(ord(c) + 1)
    while c < "\U0010fffe":
        c = chr(ord(c) + 0x10000)
        r += "|" + c
        r += "|" + chr(ord(c) + 1)

    return re.compile(r)


def get_tls_error_phrase(tls_error: Gio.TlsCertificateFlags) -> str | None:
    phrase = GIO_TLS_ERRORS.get(tls_error)
    if phrase is None:
        return GIO_TLS_ERRORS.get(Gio.TlsCertificateFlags.GENERIC_ERROR)
    return phrase


def convert_tls_error_flags(
    flags: Gio.TlsCertificateFlags | None,
) -> set[Gio.TlsCertificateFlags]:
    if not flags:
        return set()

    # If GLib ever adds more flags GIO_TLS_ERRORS have to
    # be extended, otherwise errors go unnoticed
    if Gio.TlsCertificateFlags.VALIDATE_ALL != 127:
        raise ValueError

    return set(filter(lambda error: error & flags, GIO_TLS_ERRORS.keys()))


def convert_soup_encoding(flags: int) -> set[Soup.Encoding]:
    return set(filter(lambda enc: enc & flags, SOUP_ENCODING))


def get_websocket_close_string(websocket: Soup.WebsocketConnection) -> str:
    data = websocket.get_close_data()
    code = websocket.get_close_code()

    if code is None and data is None:
        return ""
    return " Data: %s Code: %s" % (data, code)


def is_websocket_close(stanza: Node) -> bool:
    return stanza.getName() == "close" and stanza.getNamespace() == Namespace.FRAMING


def is_websocket_stream_error(stanza: Node) -> bool:
    return stanza.getName() == "error" and stanza.getNamespace() == Namespace.STREAMS


def min_version(name: str, min_version: str) -> bool:
    if name == "GLib":
        return GLIB_VERSION >= Version(min_version)
    raise ValueError("Unknown library name")


class Observable:
    def __init__(self, log_: logging.Logger | LogAdapter) -> None:
        self._log = log_
        self._frozen = False
        self._callbacks: defaultdict[str, list[Callable[..., Any]]] = defaultdict(list)

    def remove_subscriptions(self) -> None:
        self._callbacks = defaultdict(list)

    def subscribe(self, signal_name: str, func: Callable[..., Any]) -> None:
        self._callbacks[signal_name].append(func)

    def notify(self, signal_name: str, *args: Any, **kwargs: dict[str, Any]) -> None:
        if self._frozen:
            self._frozen = False
            return

        self._log.info("Signal: %s", signal_name)
        callbacks = self._callbacks.get(signal_name, [])
        for func in callbacks:
            func(self, signal_name, *args, **kwargs)


class LogAdapter(LoggerAdapter):

    def set_context(self, context: str) -> None:
        self.extra["context"] = context

    def process(self, msg: str, kwargs: Any) -> str:
        return "(%s) %s" % (self.extra["context"], msg), kwargs


def parse_websocket_uri(data: str) -> str:
    """
    Example:
    <?xml version='1.0' encoding='utf-8'?>
    <XRD xmlns='http://docs.oasis-open.org/ns/xri/xrd-1.0'>
      <Link rel='urn:xmpp:alt-connections:xbosh'
            href='https://bosh.domain.org/'/>
      <Link rel='urn:xmpp:alt-connections:websocket'
            href='wss://ws.domain.org/'/>
    </XRD>
    """
    host_meta_xml = ET.fromstring(data)
    for link in host_meta_xml.findall(f"{{{Namespace.XRD}}}Link"):
        if link.attrib.get("rel") == "urn:xmpp:alt-connections:websocket":
            href = link.attrib.get("href")
            if href is None:
                raise ValueError("No href attr found")
            return href
    raise ValueError("no websocket uri found")



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.const import TUNE_DATA
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import Node
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import TuneData
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Tune(BaseModule):

    _depends = {"publish": "PubSub"}

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pubsub_tune,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            ),
        ]

    def _process_pubsub_tune(
        self, _client: Client, _stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.TUNE:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        tune_node = item.getTag("tune", namespace=Namespace.TUNE)
        if not tune_node.getChildren():
            self._log.info("Received tune: %s - no tune set", properties.jid)
            return

        tune_dict = {}
        for attr in TUNE_DATA:
            tune_dict[attr] = tune_node.getTagData(attr)

        data = TuneData(**tune_dict)
        if data.artist is None and data.title is None:
            self._log.warning("Missing artist or title: %s %s", data, properties.jid)
            return

        pubsub_event = properties.pubsub_event._replace(data=data)
        self._log.info("Received tune: %s - %s", properties.jid, data)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def set_tune(self, data: TuneData | None):
        task = yield

        item = Node("tune", {"xmlns": Namespace.TUNE})
        if data is not None:
            data = data._asdict()
            for tag, value in data.items():
                if value is not None:
                    item.addChild(tag, payload=value)

        result = yield self.publish(Namespace.TUNE, item, id_="current")

        yield finalize(task, result)
